# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
from torch._guards import detect_fake_mode
from torch.fx.passes.infra.pass_base import PassBase, PassResult
from torch_ttnn.passes.lowering import target_wrappers
from torch_ttnn.passes.analysis.input_analysis_pass import PrimalTag

import math
import logging

import ttnn


class MultiDevicePass(PassBase):
    """Pass to introduce sharding, replication, and concatenation for multi device support.

    This pass is multi-device aware, and will return early if the input device is not a MeshDevice. Currently, this pass will introduce wrappers for sharding, replication, and concatenation that mock the shape changes that will occur with the real operations. This is used to propagate shape info throughout the computation graph. Additionally, this pass rewrites lowered constant arguments for functions that will need to change for multi-device support.

    :param device: The device that was used for compilation. Can be MeshDevice or Device.
    :param example_inputs: The inputs supplied to compilation, used for shape propagation.
    """

    def __init__(self, device, example_inputs):
        super().__init__()

        self.device = device
        self.example_inputs = example_inputs

    def replicate_to_mesh(self, node, first_node):
        with node.graph.inserting_before(first_node):
            new_node = node.graph.call_function(target_wrappers.replicate_tensor, (node,))
            node.replace_all_uses_with(new_node, delete_user_cb=lambda node: node != new_node and node.op != "output")

    def shard_to_mesh(self, node, first_node):
        batch_dimension = 0
        with node.graph.inserting_before(first_node):
            new_node = node.graph.call_function(
                target_wrappers.shard_tensor, (node, batch_dimension, self.device.get_num_devices())
            )
            node.replace_all_uses_with(new_node, delete_user_cb=lambda node: node != new_node)

    def concat_to_tensor(self, node):
        new_arg_list = []
        batch_dimension = 0
        with node.graph.inserting_before(node):
            for arg_node in node.args[0]:
                # only concat sharded nodes
                if arg_node.meta.get("is_sharded", False) is False:
                    new_arg = arg_node
                else:
                    new_arg = node.graph.call_function(
                        target_wrappers.concat_tensor, (arg_node, batch_dimension, self.device.get_num_devices())
                    )
                new_arg_list.append(new_arg)

            new_node = node.graph.create_node("output", target="output", args=(new_arg_list,))
            node.replace_all_uses_with(new_node)
        node.graph.erase_node(node)

    def rewrite_const_args(self, node):
        if node.meta.get("is_sharded", False) == False:
            return

        def get_sharded_size(concat_dim_size, node):
            num_devices = self.device.get_num_devices()
            # handle case where we have combined dimensions (in torch.view for example)
            shard_dim_size = concat_dim_size / node.meta["concat_size"]
            num_shards_per_device = math.ceil(node.meta["concat_size"] / num_devices)
            return int(shard_dim_size * num_shards_per_device)

        new_args = list(node.args)
        rewrite_arg_idx = 1
        output_dim = new_args[rewrite_arg_idx].copy()
        output_dim[node.meta["shard_dim"]] = get_sharded_size(output_dim[node.meta["shard_dim"]], node)
        new_args[rewrite_arg_idx] = output_dim

        node.args = tuple(new_args)

    def call(self, graph_module: torch.fx.GraphModule):
        if self.device.get_num_devices() < 2:
            modified = False
            return PassResult(graph_module, modified)

        node_list = list(graph_module.graph.nodes)
        first_node = [node for node in node_list if node.op != "placeholder"][0]
        for node in node_list:
            if node.op == "placeholder":
                try:
                    if (node.meta["primal_tag"] == PrimalTag.PARAMETER) or (
                        node.meta["primal_tag"] == PrimalTag.BUFFER
                    ):
                        self.replicate_to_mesh(node, first_node)
                    elif node.meta["primal_tag"] == PrimalTag.ARGUMENT:
                        self.shard_to_mesh(node, first_node)
                except Exception as e:
                    logging.error(f"Could not distribute node {node.name}: {e}")

            elif node.op == "output":
                try:
                    self.concat_to_tensor(node)
                except Exception as e:
                    logging.error(f"Could not concat node {node.name}: {e}")

            elif node.op == "call_function":
                if node.target == torch.ops.aten.view.default:
                    self.rewrite_const_args(node)
                elif node.target == torch.ops.aten._unsafe_view.default:
                    self.rewrite_const_args(node)
                elif node.target == torch.ops.aten.expand.default:
                    self.rewrite_const_args(node)

        fake_mode = detect_fake_mode(self.example_inputs)
        # propagate meta["val"]
        torch.fx.passes.fake_tensor_prop.FakeTensorProp(graph_module, mode=fake_mode).propagate(*self.example_inputs)
        # propagate meta["tensor_meta"]
        torch.fx.passes.shape_prop.ShapeProp(graph_module, fake_mode=fake_mode).propagate(*self.example_inputs)

        modified = True
        return PassResult(graph_module, modified)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
import ttnn
from torch.fx.passes.infra.pass_base import PassBase, PassResult
from torch_ttnn.passes.lowering.add_data_move_pass import is_tt_compute

from torch_ttnn import mem_utils


class MemoryPass(PassBase):
    def __init__(self, verbose=True):
        self.mm = mem_utils.MemoryManager(
            mem_utils.device_name,
            mem_utils.cores,
            mem_utils.L1_mem,
            mem_utils.circular_buffer,
            verbose,
        )
        self.op_registry = mem_utils.OpRegistry()

    # Tensor IDs allocation for ttnn ops input & output
    def assign_tensor_ids(self, gm: torch.fx.GraphModule):
        # Tensor ID allocation tracker
        last_assigned_tid = 1
        tt_compute_count = 0
        nodes = list(gm.graph.nodes)
        for node in nodes:
            if is_tt_compute(node):
                tt_compute_count += 1
                # For inputs
                for input_node in node.all_input_nodes:
                    if input_node not in self.mm.node_to_tid_map:
                        self.mm.node_to_tid_map[input_node] = last_assigned_tid
                        last_assigned_tid += 1
                # For output
                if node not in self.mm.node_to_tid_map:
                    self.mm.node_to_tid_map[node] = last_assigned_tid
                    last_assigned_tid += 1

        self.mm.log(f"Tensor IDs for nodes on device:\n")
        for key, value in self.mm.node_to_tid_map.items():
            self.mm.log(f"Node: {key} - {value}\n")
        return tt_compute_count

    def allocate(self, node: torch.fx.node.Node):
        tensor_shape, tensor_dtype = self.op_registry.get_tensor_shape_and_dtype(node)
        tensor_size = mem_utils.get_tensor_size(tensor_shape, tensor_dtype)

        assert node in self.mm.node_to_tid_map, "Tensor ID is not assigned for one of the inputs of ttnn op!"

        tid = self.mm.node_to_tid_map[node]
        self.mm.tensor_size_of[tid] = tensor_size

        start_addr = self.mm.last_assigned_addr
        self.mm.last_assigned_addr += tensor_size
        end_addr = self.mm.last_assigned_addr
        self.mm.tid_to_addr_map_in_sram[tid] = (start_addr, end_addr)

        if tid not in self.mm.tensors_on_device:
            self.mm.used_sram += tensor_size
            self.mm.free_sram -= tensor_size
            self.mm.tensors_on_device.append(tid)
        return (tid, tensor_size)

    def deallocate(self, node: torch.fx.node.Node):
        (
            tensor_shape,
            tensor_dtype,
        ) = self.op_registry.get_tensor_shape_and_dtype(node)
        tensor_size = mem_utils.get_tensor_size(
            tensor_shape,
            tensor_dtype,
        )
        self.mm.used_sram -= tensor_size
        self.mm.free_sram += tensor_size
        tid = self.mm.node_to_tid_map[node]
        self.mm.tensors_on_device.remove(tid)
        return (tid, tensor_size)

    def memory_footprint(self, gm: torch.fx.GraphModule) -> torch.fx.GraphModule:
        self.mm.log("Track memory footprint for each of the ops:\n")

        # List of already executed nodes on device
        computed_nodes = []
        # Data points containing info of on device tensors allocation & deallocation for ttnn ops
        # Useful for visualization by plotting in a chart
        data_points = {}
        # On device tensors info
        on_device_meta = []

        tt_compute_count = self.assign_tensor_ids(gm)
        self.mm.log(f"Total tt compute nodes: {tt_compute_count}\n")
        i = 0
        overflow_ops = []

        nodes = list(gm.graph.nodes)
        for node in nodes:
            on_device = False
            # Has to be a ttnn op
            if is_tt_compute(node):
                self.mm.log("---------------------------------------------------\n")
                self.mm.log(f"tt compute node numbered: {i}\n")
                i += 1
                # Exception: Full node has no input
                # Assumption: Full node outputs a tensor on device
                if node.target == ttnn.full:
                    on_device = True
                total_input_size = 0
                # Input & Output tensor IDs for ttnn op
                ttnn_ops_tids = []
                for input_node in node.all_input_nodes:
                    # If input tensor on device or coming as output from another ttnn on device op
                    if self.op_registry.is_input_tensor_on_device(input_node):
                        on_device = True
                        tid, tensor_size = self.allocate(input_node)

                        ttnn_ops_tids.append(tid)
                        total_input_size += tensor_size
                        if (tid, tensor_size) not in on_device_meta:
                            on_device_meta.append((tid, tensor_size))

                if on_device:
                    tid, output_size = self.allocate(node)

                    self.mm.ops_mem_usage[node.name] = total_input_size + output_size
                    ttnn_ops_tids.append(tid)
                    self.mm.ttnn_ops_tids[node.name] = ttnn_ops_tids
                    self.mm.tids_in_sram_at[node.name] = self.mm.tensors_on_device.copy()
                    self.mm.sram_usage_at[node.name] = self.mm.used_sram

                    if (tid, output_size) not in on_device_meta:
                        on_device_meta.append((tid, output_size))

                    data_points[(node.name, tid)] = on_device_meta.copy()

                    if self.mm.used_sram >= self.mm.usable_sram_limit:
                        overflow_ops.append(node)

                    self.mm.log(f"op execution on device: {node.name}\n")
                    self.mm.log(f"input size: {total_input_size} bytes\n")
                    self.mm.log(f"output size: {output_size} bytes\n")
                    self.mm.log(f"total SRAM usage: {self.mm.used_sram} bytes\n")

                    # This part of code is to assess whether we can swap out output tensor from device
                    node_users = list(node.users.keys())
                    # Are any users of current node a ttnn op?
                    is_follow_up_tt_compute = False
                    # Is from device op a user of the node?
                    is_follow_up_from_device = False
                    for user in node_users:
                        if is_tt_compute(user):
                            is_follow_up_tt_compute = True
                        if user.target in set([ttnn.from_device, ttnn.to_torch, ttnn.to_layout]):
                            is_follow_up_from_device = True

                    # If no follow up ttnn op using the output of current node
                    # and if one of the users is a from device operation,
                    # then swap out output tensor from device
                    if not is_follow_up_tt_compute and is_follow_up_from_device:
                        tid, output_size = self.deallocate(node)
                        on_device_meta.remove((tid, output_size))

                        self.mm.log(f"op removed from device: {node.name}\n")
                        self.mm.log(f"output size: {output_size} bytes\n")
                        self.mm.log(f"total SRAM usage: {self.mm.used_sram} bytes\n")

                    # Once current node's input & output tensor processing is over, mark it as computed
                    computed_nodes.append(node)
                    # Swap out input tensors from device if its usage is completed
                    # i.e there are no ttnn ops on device which uses the tensor
                    for input_node in node.all_input_nodes:
                        # Exception: Shape node is just like a parameter containing shape of input tensor
                        if input_node.target == ttnn.Shape:
                            continue
                        keep_on_device = False
                        input_node_users = list(input_node.users.keys())
                        for user in input_node_users:
                            if user not in computed_nodes and is_tt_compute(user):
                                keep_on_device = True
                        if keep_on_device is False:
                            tid, tensor_size = self.deallocate(input_node)
                            on_device_meta.remove((tid, tensor_size))

                            self.mm.log(f"input tensor removed from device: {tensor_size} bytes\n")
                            self.mm.log(f"total SRAM usage: {self.mm.used_sram} bytes\n")

                    # data_points[(node.name, "from_device")] = on_device_meta.copy()

        self.mm.set_peak_sram_usage()

        self.mm.log(f"Tensor IDs to address map in SRAM:\n")
        self.mm.log(f"{self.mm.tid_to_addr_map_in_sram}\n")
        self.mm.log(f"----------------------------------------------------------------\n")
        self.mm.log(f"These ttnn ops overflow the SRAM buffer:\n")
        for op in overflow_ops:
            self.mm.log(f"{op}, ")

        self.mm.log(f"\nData captured for plotting on a chart:\n")
        for key, value in data_points.items():
            self.mm.log(f"{key}: {value}\n")

        # Convert dictionary keys to strings
        data_str_keys = {str(key): value for key, value in data_points.items()}
        self.mm.data_points = data_str_keys

        return gm

    def call(self, gm: torch.fx.GraphModule):
        gm = self.memory_footprint(gm)
        return PassResult(gm, True)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
from torch_ttnn import fx_graphviz

from torch.fx.passes.infra.pass_base import PassBase, PassResult


class GraphvizPass(PassBase):
    def __init__(self, filename):
        super().__init__()
        self.filename = filename

    def call(self, gm: torch.fx.GraphModule):
        fx_graphviz.to_svg(gm.graph, self.filename)
        return PassResult(gm, False)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import inspect
import pickle
import torch
import ttnn
from torch_ttnn.passes.analysis.graph_module_analysis_pass import ModelType
from torch_ttnn.passes.analysis.input_analysis_pass import PrimalTag
from torch_ttnn.utils import GraphCleanup, TtnnDevice, TtnnBfloat16, TtnnUint32, TtnnTileLayout, get_dtype
from operator import getitem

from torch.fx.passes.infra.pass_base import PassBase, PassResult
from . import target_wrappers


def mark_nodes_invariant(nodes):
    seen_nodes = set()

    def mark_descendants_varied(node_list):
        while len(node_list) > 0:
            node = node_list.pop(-1)
            if node in seen_nodes:
                continue

            node.meta["iteration_invariant"] = False
            seen_nodes.add(node)

            node_list += node.users

    for node in nodes:
        node.meta["iteration_invariant"] = True
    for node in filter(lambda n: n.op == "placeholder" and n.meta.get("primal_tag") == PrimalTag.ARGUMENT, nodes):
        # mark all descendents as not iteration_invariant
        mark_descendants_varied([node])
    for node in filter(lambda n: n.op in ["output", "get_attr", "root"], nodes):
        # make sure all outputs, roots, and get_attr nodes are marked as variable
        mark_descendants_varied([node])


class NodeMover:
    def __init__(self, graph):
        self.graph = graph
        self.node_to_run_once_idx = dict()
        self.node_to_returned_idx = dict()
        self.run_once_inputs = []

        nodes = self.graph.nodes
        self.first_node = next(
            filter(
                lambda node: node.op != "placeholder"
                and (node.op != "call_function" or node.target != ttnn.from_torch),
                nodes,
            )
        )

    def move(self, node):
        assert node.meta.get("iteration_invariant") == True
        if node.op in ["placeholder", "get_attr", "output", "root"]:
            return 0

        self.node_to_run_once_idx[node] = len(self.node_to_run_once_idx)
        # if any users are variable, return too
        should_return = False
        if any([n.meta.get("iteration_invariant", False) == False for n in node.users]):
            should_return = True
            self.node_to_returned_idx[node] = len(self.node_to_returned_idx)

        run_once_idx = target_wrappers.RunOnceIdx(self.node_to_run_once_idx[node])
        run_once_idx = pickle.dumps(run_once_idx)
        node.replace_all_uses_with(
            run_once_idx, delete_user_cb=lambda node: node.meta.get("iteration_invariant") == True
        )

        if hasattr(node.target, "python_fully_qualified_name"):
            fn_call = node.target.python_fully_qualified_name
        elif node.target.__module__ == "torch_ttnn.passes.lowering.target_wrappers":
            fn_call = node.target.__qualname__
        elif node.target.__module__ == "torch._ops.aten":
            fn_call = f"torch.ops.aten.{node.target.__name__}"
        elif node.target.__module__ == "ttnn.distributed.distributed":
            fn_call = f"ttnn.{node.target.__name__}"
        elif inspect.isbuiltin(node.target):
            fn_call = node.target.__name__
        else:
            fn_call = f"{node.target.__module__}.{node.target.__qualname__}"

        self.run_once_inputs.append((should_return, fn_call, node.args, node.kwargs))
        return 1

    def move_conv(self, node):
        new_args = list(node.args)
        modified_count = 0

        # rewrite input because otherwise it is not defined when run_once is called
        input_node = node.args[0]
        tensor_shape = list(input_node.meta["val"].shape)
        spec_dtype = TtnnBfloat16
        if get_dtype(input_node) in [torch.int32, torch.int64]:
            spec_dtype = TtnnUint32
        # currently, all conv inputs are tiled and on device. If this changes, may need further analysis here
        mocked_input_node = (
            False,
            "ttnn.ones",
            (),
            {"shape": tensor_shape, "device": TtnnDevice(), "dtype": spec_dtype(), "layout": TtnnTileLayout()},
        )
        mocked_input_node_idx = len(self.run_once_inputs)
        self.run_once_inputs.append(mocked_input_node)
        input_node_idx = target_wrappers.RunOnceIdx(mocked_input_node_idx)
        new_args[0] = pickle.dumps(input_node_idx)

        should_preprocess = False

        # rewrite weight and optional bias as RunOnceIdx values
        weight_node = node.args[1]
        if weight_node.meta.get("iteration_invariant", False):
            should_preprocess = True
            weight_node_idx = target_wrappers.RunOnceIdx(self.node_to_run_once_idx[weight_node])
            new_args[1] = pickle.dumps(weight_node_idx)
            modified_count += 1

        bias_node = node.args[2]
        if bias_node is not None and bias_node.meta.get("iteration_invariant", False):
            should_preprocess = True
            bias_node_idx = target_wrappers.RunOnceIdx(self.node_to_run_once_idx[bias_node])
            new_args[2] = pickle.dumps(bias_node_idx)
            modified_count += 1

        if not should_preprocess:
            # only preprocess if weight or optional bias are iteration invariant
            return 0

        new_args = tuple(new_args)

        # update kwargs to return weights and bias
        new_kwargs = node.kwargs.copy()
        new_kwargs["return_weights_and_bias"] = True

        should_return = False
        fn_call = "conv"

        self.run_once_inputs.append((should_return, fn_call, new_args, new_kwargs))
        return modified_count

    def rewrite_users(self, ttnn_inputs):
        with self.graph.inserting_before(self.first_node):
            for node, idx in self.node_to_returned_idx.items():
                getitem_call = self.graph.call_function(getitem, (ttnn_inputs, idx))
                getitem_call.meta["is_cached"] = True
                # replace all varying users with getitem call
                node.replace_all_uses_with(
                    getitem_call, delete_user_cb=lambda node: node.meta.get("iteration_invariant") == False
                )

    def create_load_once(self):
        if len(self.run_once_inputs) == 0:
            return

        global run_once_count
        target_wrappers.run_once_count = 0

        if len(self.run_once_inputs) > 0:
            with self.graph.inserting_before(self.first_node):
                ttnn_inputs = self.graph.call_function(
                    target_wrappers.run_once,
                    tuple(self.run_once_inputs),
                )

            self.rewrite_users(ttnn_inputs)


class LoadOncePass(PassBase):
    """Pass that moves as many operations as possible into the load_once function so they will be cached for later runs.

    This pass only runs for end to end inference models.

    :param is_end_to_end: Whether the input GraphModule is converted end to end.
    """

    def __init__(self, is_end_to_end, load_params_once):
        self.is_end_to_end = is_end_to_end
        self.load_params_once = load_params_once

    def call(self, gm: torch.fx.GraphModule):
        if (not self.is_end_to_end) or (not self.load_params_once) or gm.meta.get("graph_type") != ModelType.INFERENCE:
            modified = False
            return PassResult(gm, modified)

        modifications_count = 0

        # first make sure all nodes are marked as constant or variable (infected by arguments)
        nodes = list(gm.graph.nodes)
        mark_nodes_invariant(nodes)

        # move all constant nodes into run_once and rewrite users
        node_mover = NodeMover(gm.graph)
        for constant_node in filter(lambda n: n.meta.get("iteration_invariant", False), nodes):
            modifications_count += node_mover.move(constant_node)

        # add convolutions to preprocess weights and biases
        for conv_node in filter(lambda n: n.op == "call_function" and n.target == target_wrappers.conv, nodes):
            modifications_count += node_mover.move_conv(conv_node)

        node_mover.create_load_once()

        modified = modifications_count > 0
        GraphCleanup(gm)
        return PassResult(gm, modified)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0

import torch
import ttnn
import operator
from torch.fx import Node
from torch_ttnn.passes.patterns.pattern_matcher_base import PatternMatcherBase
from typing import List, Tuple


class AttentionPatterns(PatternMatcherBase):
    """Pattern matcher for attention operations."""

    def match_pattern(self) -> List[Tuple[torch.fx.Node, ...]]:
        """
        Match the attention pattern:
        linear -> view -> reshape -> permute -> view -> matmul(q, k) -> view -> softmax -> view -> matmul(softmax, v) -> view -> view -> permute -> reshape -> view

        Returns a list of tuples: (q_linear, q_view1, q_reshape, q_permute, q_view2,
                                 k_linear, k_view1, k_reshape, k_permute, k_transpose, k_view2,
                                 v_linear, v_view1, v_reshape, v_permute, v_view2,
                                 matmul_qk, view1, softmax, view2, matmul_sv, view3, permute, reshape, view5,
                                 attn_mask, scale)
        """
        matches = []

        # Find all potential QKV candidates
        query_candidates = self._find_qkv_candidates(False, 0)
        key_candidates = self._find_qkv_candidates(True, 1)
        value_candidates = self._find_qkv_candidates(False, 1)

        # For each query candidate, try to find matching key and value candidates
        for q_candidate in query_candidates:
            q_linear, q_view1, q_reshape, q_permute, q_view2, q_matmul = q_candidate

            # Find matching key candidate where the matmul connects q and k
            for k_candidate in key_candidates:
                k_linear, k_view1, k_reshape, k_permute, k_transpose, k_view2, k_matmul = k_candidate

                # check that the weights have the same shape
                if q_linear.args[1].meta["val"].shape != k_linear.args[1].meta["val"].shape:
                    continue

                # Check if this is the QK matmul
                if k_matmul is not q_matmul:
                    continue

                # Check if QK matmul output goes to a view
                view1 = self._find_exclusive_user_of_type(q_matmul, ttnn.experimental.view)
                if not view1:
                    continue

                # Check if view output goes to softmax and get the mask if it exists
                softmax = self._find_exclusive_user_of_type(view1, ttnn.transformer.attention_softmax)
                if not softmax:
                    continue

                # Get attention mask if it exists (it would be an additional argument to softmax)
                attn_mask = softmax.kwargs.get("attention_mask")
                # Get scale if it exists in kwargs
                head_size = softmax.kwargs.get("head_size")

                # If head_size is not provided, we cannot proceed
                # at least for now...
                if not head_size:
                    continue

                scale = 1.0 / (head_size**0.5) if head_size is not None else None

                # Check if softmax output goes to a view
                view2 = self._find_exclusive_user_of_type(softmax, ttnn.experimental.view)
                if not view2:
                    continue

                # Find matching value candidate
                for v_candidate in value_candidates:
                    v_linear, v_view1, v_reshape, v_permute, v_view2, v_matmul = v_candidate

                    # check that the weights have the same shape
                    if q_linear.args[1].meta["val"].shape != v_linear.args[1].meta["val"].shape:
                        continue

                    # Check if view output goes to matmul with value
                    matmul_sv = self._find_exclusive_user_of_type(view2, ttnn.matmul)
                    if not matmul_sv or matmul_sv is not v_matmul:
                        continue

                    # Check for additional operations after matmul_sv
                    view3 = self._find_exclusive_user_of_type(matmul_sv, ttnn.experimental.view)
                    if not view3:
                        continue

                    permute = self._find_exclusive_user_of_type(view3, ttnn.permute)
                    if not permute:
                        continue

                    reshape = self._find_exclusive_user_of_type(permute, ttnn.reshape)
                    if not reshape:
                        continue

                    view5 = self._find_exclusive_user_of_type(reshape, ttnn.experimental.view)
                    if not view5:
                        continue

                    # We found a complete attention pattern
                    matches.append(
                        (
                            q_linear,
                            q_view1,
                            q_reshape,
                            q_permute,
                            q_view2,
                            k_linear,
                            k_view1,
                            k_reshape,
                            k_permute,
                            k_transpose,
                            k_view2,
                            v_linear,
                            v_view1,
                            v_reshape,
                            v_permute,
                            v_view2,
                            q_matmul,
                            view1,
                            softmax,
                            view2,
                            matmul_sv,
                            view3,
                            permute,
                            reshape,
                            view5,
                            attn_mask,
                            scale,
                        )
                    )

        return matches

    # Hack! TTNN does not support concat with tensors of 1D
    # https://github.com/tenstorrent/tt-metal/issues/20205
    # bias is 1D tensor, so we need to make it 2D
    def _hack_1D_bias(self, bias_node):
        """Reshape bias node to 2D Shape

        Args:
            bias_node: The bias node to reshape

        Returns:
            The reshaped node
        """
        new_bias_node = self.gm.graph.call_function(
            ttnn.experimental.view,
            args=(bias_node, (1, -1)),
            kwargs={},
        )

        return new_bias_node

    def replace_pattern(self, matches: List[Tuple[torch.fx.Node, ...]]) -> None:
        for match in matches:
            (
                q_linear,
                q_view1,
                q_reshape,
                q_permute,
                q_view2,
                k_linear,
                k_view1,
                k_reshape,
                k_permute,
                k_transpose,
                k_view2,
                v_linear,
                v_view1,
                v_reshape,
                v_permute,
                v_view2,
                matmul_qk,
                view1,
                softmax,
                view2,
                matmul_sv,
                view3,
                permute,
                reshape,
                view5,
                attn_mask,
                scale,
            ) = match

            head_size = softmax.kwargs.get("head_size")
            num_heads = q_view1.meta["val"].shape[-1] // head_size

            with self.gm.graph.inserting_before(view5):
                transposed_q_weights_node = self.gm.graph.call_function(
                    ttnn.transpose,
                    args=(q_linear.args[1], -2, -1),
                )

                transposed_k_weights_node = self.gm.graph.call_function(
                    ttnn.transpose,
                    args=(k_linear.args[1], -2, -1),
                    kwargs={},
                )

                transposed_v_weights_node = self.gm.graph.call_function(
                    ttnn.transpose,
                    args=(v_linear.args[1], -2, -1),
                    kwargs={},
                )

                # Lets concatenate the qkv tensors for weights and biases
                concatenated_qkv_weights_node = self.gm.graph.call_function(
                    ttnn.concat,
                    args=([transposed_q_weights_node, transposed_k_weights_node, transposed_v_weights_node], -1),
                    kwargs={},
                )

                # _find_qkv_pattern ensures that the bias is not None
                # this is not entirely true, because the bias is optional
                # but since I want to control the replacement, I will make it a requirement
                # for now.
                q_bias = q_linear.kwargs["bias"]
                k_bias = k_linear.kwargs["bias"]
                v_bias = v_linear.kwargs["bias"]

                # Reshape biases if needed
                q_bias_node = self._hack_1D_bias(q_bias)
                k_bias_node = self._hack_1D_bias(k_bias)
                v_bias_node = self._hack_1D_bias(v_bias)

                # concatenate the biases
                concatenated_qkv_biases_node = self.gm.graph.call_function(
                    ttnn.concat,
                    args=([q_bias_node, k_bias_node, v_bias_node], -1),
                    kwargs={},
                )

                qvk_node = self.gm.graph.call_function(
                    ttnn.linear,
                    args=(
                        q_linear.args[0],
                        concatenated_qkv_weights_node,
                    ),
                    kwargs={"bias": concatenated_qkv_biases_node},
                )

                shape = list(q_view1.meta["val"].shape)
                shape[-1] = shape[-1] * 3

                linear_view_node = self.gm.graph.call_function(
                    ttnn.experimental.view,
                    args=(qvk_node, shape),
                    kwargs={},
                )

                # Time to split the qvk into q, k, v
                split_qkv_node = self.gm.graph.call_function(
                    ttnn.transformer.split_query_key_value_and_split_heads,
                    args=(linear_view_node, None),
                    kwargs={"num_heads": num_heads, "transpose_key": True},
                )

                get_q_node = self.gm.graph.call_function(
                    operator.getitem,
                    args=(split_qkv_node, 0),
                    kwargs={},
                )

                get_k_node = self.gm.graph.call_function(
                    operator.getitem,
                    args=(split_qkv_node, 1),
                    kwargs={},
                )

                get_v_node = self.gm.graph.call_function(
                    operator.getitem,
                    args=(split_qkv_node, 2),
                    kwargs={},
                )

                attention_scores_node = self.gm.graph.call_function(
                    ttnn.matmul,
                    args=(get_q_node, get_k_node),
                )

                attention_probabilities_node = self.gm.graph.call_function(
                    ttnn.transformer.attention_softmax,
                    args=(attention_scores_node,),
                    kwargs={
                        "attention_mask": attn_mask,
                        "head_size": head_size,
                    },
                )

                matmul_sv_node = self.gm.graph.call_function(
                    ttnn.matmul,
                    args=(attention_probabilities_node, get_v_node),
                    kwargs={},
                )

                concatenated_head_node = self.gm.graph.call_function(
                    ttnn.transformer.concatenate_heads,
                    args=(matmul_sv_node,),
                    kwargs={},
                )

            view5.replace_all_uses_with(concatenated_head_node)

            # Remove old nodes in reverse order
            nodes_to_remove = [
                view5,
                reshape,
                permute,
                view3,
                matmul_sv,
                view2,
                softmax,
                view1,
                matmul_qk,
                q_view2,
                q_permute,
                q_reshape,
                k_view2,
                k_transpose,
                k_permute,
                k_reshape,
                v_view2,
                v_permute,
                v_reshape,
                q_view1,
                k_view1,
                v_view1,
                q_linear,
                k_linear,
                v_linear,
            ]
            self.safe_remove_nodes(nodes_to_remove)

        return matches

    def _find_qkv_pattern(self, linear, require_transpose=False, matmul_arg_idx=0):
        """Common helper method for finding query/key/value patterns.

        Args:
            linear: The linear node to start from
            require_transpose: Whether to require a transpose operation
            matmul_arg_idx: Which argument of matmul should be the view2 node (0 for query, 1 for key/value)

        Returns:
            Tuple of nodes if pattern matches, None otherwise
        """

        # Even though the bias is optional, I want to restrict the pattern to only include nodes with a bias
        if linear.kwargs["bias"] is None:
            return None

        view1 = self._find_exclusive_user_of_type(linear, ttnn.experimental.view)
        if not view1:
            return None

        reshape = self._find_exclusive_user_of_type(view1, ttnn.reshape)
        if not reshape:
            return None

        permute = self._find_exclusive_user_of_type(reshape, ttnn.permute)
        if not permute:
            return None

        if require_transpose:
            transpose = self._find_exclusive_user_of_type(permute, ttnn.transpose)
            if not transpose:
                return None
            view2 = self._find_exclusive_user_of_type(transpose, ttnn.experimental.view)
        else:
            view2 = self._find_exclusive_user_of_type(permute, ttnn.experimental.view)

        if not view2:
            return None

        matmul = self._find_exclusive_user_of_type(view2, ttnn.matmul)
        if not matmul or matmul.args[matmul_arg_idx] is not view2:
            return None

        if require_transpose:
            return (linear, view1, reshape, permute, transpose, view2, matmul)
        return (linear, view1, reshape, permute, view2, matmul)

    def _find_qkv_candidates(self, require_transpose, matmul_arg_idx):
        key_candidates = []
        linear_nodes = self._find_nodes_of_type(ttnn.linear)
        for linear in linear_nodes:
            candidate = self._find_qkv_pattern(
                linear, require_transpose=require_transpose, matmul_arg_idx=matmul_arg_idx
            )

            if candidate:
                key_candidates.append(candidate)
        return key_candidates



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0

import torch
import ttnn
from torch.fx import Node
from torch_ttnn.passes.patterns.pattern_matcher_base import PatternMatcherBase
from typing import List, Tuple


class LinearPatterns(PatternMatcherBase):
    """Pattern matcher for linear operations with optional activation and views."""

    def match_pattern(self) -> List[Tuple[torch.fx.Node, ...]]:
        """
        Match the linear pattern: arg -> transpose -> matmul -> add(arg_2) -> view -> [activation] -> [view2]
        where activation can be gelu, relu, or silu (optional), followed by an optional second view.
        Returns a list of tuples containing the matched nodes.

        The expected output will be:
        ttnn.linear(input_tensor, weight, bias, activation)
        """
        matches = []

        # Find all transpose nodes
        transpose_nodes = self._find_nodes_of_type(ttnn.transpose)

        for transpose in transpose_nodes:
            # The input to transpose can be any node (arg)
            arg = transpose.args[0] if len(transpose.args) > 0 else None
            if arg is None:
                continue

            # Find matmul operation using transpose
            matmul = self._find_exclusive_user_of_type(transpose, ttnn.matmul)
            if not matmul or not self._is_node_in_args(transpose, matmul):
                continue

            # Look for add operation that uses matmul and any other node as bias (arg_2)
            add = None
            arg_2 = None
            for user in matmul.users:
                if self._is_ttnn_op(user, ttnn.add):
                    # Find the other argument to add (not matmul)
                    for possible_arg_2 in user.args:
                        if possible_arg_2 is not matmul:
                            add = user
                            arg_2 = possible_arg_2
                            break
                if add is not None:
                    break

            if add is None or arg_2 is None:
                continue

            # in ttnn.linear, B can't be batched and have bias
            # in this case, B is the one that is going to be transposed
            shape = arg.meta["tensor_meta"].shape
            batched = False
            if len(shape) > 2:
                dims_to_check = shape[: 1 if len(shape) == 3 else 2]
                for dim in dims_to_check:
                    batched |= dim != 1

            if batched:
                continue

            # Create a mapping of activation function types to their string names
            activation_map = {ttnn.gelu: "gelu", ttnn.relu: "relu", ttnn.silu: "silu"}

            potential_replaceable_view = self._find_exclusive_user_of_type(add, ttnn.experimental.view)
            if potential_replaceable_view is not None:
                # Default activation to None
                activation = (None, None)

                # Check each activation type from the map
                for act_func, act_name in activation_map.items():
                    act_node = self._find_single_user_of_type(potential_replaceable_view, act_func)
                    if act_node:
                        activation = (act_name, act_node)
                        break

                # If no activation was found, set view to None
                if activation[0] is None:
                    potential_replaceable_view = None

                matches.append((transpose, matmul, add, potential_replaceable_view, activation))
        return matches

    def replace_pattern(self, matches: List[Tuple[torch.fx.Node, ...]]) -> None:
        for match in matches:
            transpose, matmul, add, view, activation = match

            bias_arg = add.args[1] if add.args[0] is matmul else add.args[0]
            # If there is no activation, keep the last view node
            last_node = activation[1] if activation[0] else add
            potential_activation = activation[0] if activation[0] else None
            with self.gm.graph.inserting_after(last_node):
                fused_node = self.gm.graph.call_function(
                    ttnn.linear,
                    args=(matmul.args[0], transpose.args[0]),
                    kwargs={
                        "transpose_b": True,
                        "bias": bias_arg,
                        "activation": potential_activation,
                    },
                )

                # Connect output to the next node
                (last_node).replace_all_uses_with(fused_node)

            # Remove old nodes in reverse order to handle dependencies
            nodes_to_remove = [
                activation[1] if potential_activation else None,
                view if activation[0] else None,
                add,
                matmul,
                transpose,
            ]

            self.safe_remove_nodes(nodes_to_remove)

        self.gm.graph.lint()
        self.gm.recompile()
        return matches



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import graphviz
import torch
import ttnn
import os
import math
from collections import defaultdict


def _tensor_weight(t):
    s = t.shape
    if s is None:
        return 1

    result = 1
    for d in s:
        if d is None:
            continue
        else:
            result *= d

    return result


def _tensor_width(t):
    w = _tensor_weight(t)
    # Strange, some tensor has 0 element
    if w <= 0:
        w = 1
    l = math.log10(w) * 2
    return max(l, 4)


def _tensor_label(t):
    return str(t.shape)


def node_name(node):
    # FX graph node
    # The node name is uquie and can be used as a key
    if isinstance(node, torch.fx.node.Node):
        return node.name

    # Other Python Object, like "device" or constant value
    # TODO(yoco): not all device are the same, maybe we should use a different name
    return str(node) + hex(id(node))


def node_label(node):
    # If it is not a FX node, just return the string
    if not isinstance(node, torch.fx.node.Node):
        return str(node)

    # If it is a FX call function
    if node.op == "call_function":
        # A func wrapper with "_name" attribute
        if hasattr(node.target, "_name"):
            return node.target._name
        # Normal function: __module__.__name__
        return f"{node.target.__module__}.{node.target.__name__}"

    # call_module, call_method, output, and placeholder
    return str(node.op + "\n" + node.name)


def node_fillcolor(node):
    # Green for supported ops
    if node.target in set(
        [
            ttnn.add,
            ttnn.matmul,
            ttnn.sub,
            ttnn.mul,
            ttnn.softmax,
            ttnn.tanh,
            ttnn.reshape,
            ttnn.permute,
        ]
    ):
        return "#aaffaa"

    # Yellow for device ops
    if node.target in [ttnn.to_device, ttnn.from_device]:
        return "#ffffaa"

    # Orange for torch tensor conversion ops
    if node.target in [ttnn.from_torch, ttnn.to_torch]:
        return "#ffddaa"

    # Orange for torch tensor conversion ops
    if node.target in [ttnn.to_layout]:
        return "#ffeeaa"

    # Red for call_function (unknown ops)
    if node.op == "call_function":
        return "#ffaaaa"

    # Gray for others
    return "#dddddd"


def to_port(in_size, it_idx):
    port_table = [
        [],
        [""],
        [":nw", ":ne"],
        [":nw", ":n", ":ne"],
        [":w", ":nw", ":n", ":ne"],
        [":w", ":nw", ":n", ":ne", ":e"],
        [":nw", ":nw", ":n", ":n", ":ne", ":ne"],
    ]
    if in_size > 6:
        return ""
    result = port_table[in_size][it_idx]
    return result


def to_svg(g: torch.fx.Graph, filename: str):
    # Setup dot
    dot = graphviz.Digraph()
    dot.graph_attr["ranksep"] = "0.4"
    dot.node_attr["style"] = "rounded,filled"
    dot.node_attr["height"] = "0.3"
    dot.node_attr["shape"] = "box"
    dot.node_attr["fillcolor"] = "#dddddd"
    dot.edge_attr["color"] = "#77777777"

    # setup nodes
    map_node_idx = dict()
    map_idx_node = dict()
    for op_idx, node in enumerate(g.nodes):
        map_node_idx[node] = op_idx
        map_idx_node[op_idx] = node
        dot.node(node_name(node), label=node_label(node), fillcolor=node_fillcolor(node))

    # setup edges
    edge_color_table = [
        "#77ffff77",
        "#ff77ff77",
        "#7777ff77",
        "#ff777777",
        "#77ff7777",
    ]

    # A table of non-fx.node, like constant or other object that has been used.
    # For those objects, we will use a different name for each of them.
    var_obj_cnt = defaultdict(int)
    for node in g.nodes:
        # If the node is output, the "args" is a tuple like ((v1, v2),).
        # else, the "args" is a tuple like (v1, v2).
        # It is very stange, but we need to handle it.
        if node.op == "output":
            in_nodes = node.args[0]
        else:
            in_nodes = node.args

        for idx, in_node in enumerate(in_nodes):
            # For each input:
            # If it is a fx.node, use the node name.
            # else create a duplicated visual node for the input.
            if isinstance(in_node, torch.fx.node.Node):
                src_name = node_name(in_node)
            else:
                serial_num = var_obj_cnt[in_node]
                src_name = f"{node_name(in_node)}_{serial_num}"
                src_label = node_label(in_node)
                var_obj_cnt[in_node] += 1
                dot.node(src_name, label=src_label)

            # Draw the edge
            dst_name = node_name(node) + to_port(len(in_nodes), idx)
            edge_color = edge_color_table[(id(node) + idx) % 5]
            dot.edge(
                src_name,
                dst_name,
                color=edge_color,
                penwidth=str(4),
            )

    # Write .dot & .svg
    dot.save(f"{filename}.dot")
    os.system(f'sfdp -x -Goverlap=scale -Tsvg "{filename}.dot" > "{filename}.dot.svg"')



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
# THIS FILE IS AUTOGENERATED BY tools/generate_guard_function_from_input_var_metric.py
# DO NOT EDIT THIS FILE MANUALLY
# YOU CAN DEBUG THESE INPUT VAR VIA tests/autogen-op/ALL
import torch
import torch_ttnn.metrics as metrics
from functools import partial


aten__log_softmax_default_blocklist = [["Tensor<[19, 256008]> self = ?", "int dim = 1", "bool half_to_float = False"]]
aten_native_layer_norm_default_blocklist = [
    [
        "Tensor<[1, 9, 4096]> input = ?",
        "List[int] normalized_shape = [4096]",
        "Optional[Tensor]<[4096]> weight = ?",
        "Optional[Tensor]<[4096]> bias = ?",
        "float eps = 1e-12",
    ],
    [
        "Tensor<[1, 7, 4544]> input = ?",
        "List[int] normalized_shape = [4544]",
        "Optional[Tensor]<[4544]> weight = ?",
        "Optional[Tensor]<[4544]> bias = ?",
        "float eps = 1e-05",
    ],
]
aten_convolution_default_blocklist = [
    # TODO(#385): Guard and fallback (likely) OOM cases
    [
        "Tensor<[1, 3, 224, 224]> input = ?",
        "Tensor<[768, 3, 32, 32]> weight = ?",
        "Optional[Tensor] bias = ?",
        "List[int] stride = [32, 32]",
        "List[int] padding = [0, 0]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 3, 224, 224]> input = ?",
        "Tensor<[1024, 3, 32, 32]> weight = ?",
        "Optional[Tensor]<[1024]> bias = ?",
        "List[int] stride = [32, 32]",
        "List[int] padding = [0, 0]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 64, 480, 640]> input = ?",
        "Tensor<[64, 64, 3, 3]> weight = ?",
        "Optional[Tensor]<[64]> bias = ?",
        "List[int] stride = [1, 1]",
        "List[int] padding = [1, 1]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 64, 480, 640]> input = ?",
        "Tensor<[1, 64, 3, 3]> weight = ?",
        "Optional[Tensor]<[1]> bias = ?",
        "List[int] stride = [1, 1]",
        "List[int] padding = [1, 1]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 3, 384, 512]> input = ?",
        "Tensor<[768, 3, 32, 32]> weight = ?",
        "Optional[Tensor]<[768]> bias = ?",
        "List[int] stride = [32, 32]",
        "List[int] padding = [0, 0]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 3, 800, 1088]> input = ?",
        "Tensor<[64, 3, 7, 7]> weight = ?",
        "Optional[Tensor] bias = ?",
        "List[int] stride = [2, 2]",
        "List[int] padding = [3, 3]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 256, 100, 136]> input = ?",
        "Tensor<[819, 256, 3, 3]> weight = ?",
        "Optional[Tensor]<[819]> bias = ?",
        "List[int] stride = [1, 1]",
        "List[int] padding = [1, 1]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 256, 50, 68]> input = ?",
        "Tensor<[819, 256, 3, 3]> weight = ?",
        "Optional[Tensor]<[819]> bias = ?",
        "List[int] stride = [1, 1]",
        "List[int] padding = [1, 1]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 3, 720, 1280]> input = ?",
        "Tensor<[64, 3, 7, 7]> weight = ?",
        "Optional[Tensor] bias = ?",
        "List[int] stride = [2, 2]",
        "List[int] padding = [3, 3]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 256, 180, 320]> input = ?",
        "Tensor<[512, 256, 1, 1]> weight = ?",
        "Optional[Tensor] bias = ?",
        "List[int] stride = [2, 2]",
        "List[int] padding = [0, 0]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 3, 224, 224]> input = ?",
        "Tensor<[768, 3, 32, 32]> weight = ?",
        "Optional[Tensor]<[768]> bias = ?",
        "List[int] stride = [32, 32]",
        "List[int] padding = [0, 0]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 3, 224, 224]> input = ?",
        "Tensor<[768, 3, 16, 16]> weight = ?",
        "Optional[Tensor]<[768]> bias = ?",
        "List[int] stride = [16, 16]",
        "List[int] padding = [0, 0]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 3, 512, 672]> input = ?",
        "Tensor<[192, 3, 16, 16]> weight = ?",
        "Optional[Tensor]<[192]> bias = ?",
        "List[int] stride = [16, 16]",
        "List[int] padding = [0, 0]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 3, 518, 518]> input = ?",
        "Tensor<[1280, 3, 14, 14]> weight = ?",
        "Optional[Tensor]<[1280]> bias = ?",
        "List[int] stride = [14, 14]",
        "List[int] padding = [0, 0]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 3, 224, 224]> input = ?",
        "Tensor<[1024, 3, 16, 16]> weight = ?",
        "Optional[Tensor]<[1024]> bias = ?",
        "List[int] stride = [16, 16]",
        "List[int] padding = [0, 0]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 528, 192, 192]> input = ?",
        "Tensor<[528, 264, 3, 3]> weight = ?",
        "Optional[Tensor] bias = ?",
        "List[int] stride = [2, 2]",
        "List[int] padding = [1, 1]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 2",
    ],
    [
        "Tensor<[1, 1056, 96, 96]> input = ?",
        "Tensor<[1056, 264, 3, 3]> weight = ?",
        "Optional[Tensor] bias = ?",
        "List[int] stride = [2, 2]",
        "List[int] padding = [1, 1]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 4",
    ],
    [
        "Tensor<[1, 2904, 48, 48]> input = ?",
        "Tensor<[2904, 264, 3, 3]> weight = ?",
        "Optional[Tensor] bias = ?",
        "List[int] stride = [2, 2]",
        "List[int] padding = [1, 1]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 11",
    ],
    [
        "Tensor<[1, 816, 19, 19]> input = ?",
        "Tensor<[816, 1, 5, 5]> weight = ?",
        "Optional[Tensor] bias = ?",
        "List[int] stride = [1, 1]",
        "List[int] padding = [2, 2]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 816",
    ],
    [
        "Tensor<[1, 816, 23, 23]> input = ?",
        "Tensor<[816, 1, 5, 5]> weight = ?",
        "Optional[Tensor] bias = ?",
        "List[int] stride = [2, 2]",
        "List[int] padding = [0, 0]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 816",
    ],
    [
        "Tensor<[1, 960, 24, 24]> input = ?",
        "Tensor<[960, 1, 5, 5]> weight = ?",
        "Optional[Tensor] bias = ?",
        "List[int] stride = [1, 1]",
        "List[int] padding = [2, 2]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 960",
    ],
    [
        "Tensor<[1, 960, 27, 27]> input = ?",
        "Tensor<[960, 1, 5, 5]> weight = ?",
        "Optional[Tensor] bias = ?",
        "List[int] stride = [2, 2]",
        "List[int] padding = [0, 0]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 960",
    ],
    [
        "Tensor<[1, 768, 3000]> input = ?",
        "Tensor<[768, 768, 3]> weight = ?",
        "Optional[Tensor]<[768]> bias = ?",
        "List[int] stride = [2]",
        "List[int] padding = [1]",
        "List[int] dilation = [1]",
        "bool transposed = False",
        "List[int] output_padding = [0]",
        "int groups = 1",
    ],
    [
        "Tensor<[1, 320, 64, 64]> input = ?",
        "Tensor<[320, 320, 3, 3]> weight = ?",
        "Optional[Tensor]<[320]> bias = ?",
        "List[int] stride = [1, 1]",
        "List[int] padding = [1, 1]",
        "List[int] dilation = [1, 1]",
        "bool transposed = False",
        "List[int] output_padding = [0, 0]",
        "int groups = 1",
    ],
]

aten_argmax_default_blocklist = [
    ["Tensor<[2, 7]> self = ?", "Optional[int] dim = -1"],
    ["Tensor<[1, 7]> self = ?", "Optional[int] dim = -1"],
    ["Tensor<[1, 51865]> self = ?", "Optional[int] dim = -1"],
]


def get_inputs(node):
    node_inputs = metrics.collect_input_variation_from_node(node)
    if type(node_inputs) == metrics.InputVariation:
        return node_inputs.inputs
    elif type(node_inputs) == metrics.ConvertedInput:
        return node_inputs.original_input_variation.inputs
    else:
        return None


def guard_aten(blocklist, node):
    inputs = get_inputs(node)
    inputs_str = [str(i) for i in inputs]
    if inputs_str in blocklist:
        return False
    return True


GUARD = {
    torch.ops.aten._log_softmax.default: partial(guard_aten, aten__log_softmax_default_blocklist),
    torch.ops.aten.native_layer_norm.default: partial(guard_aten, aten_native_layer_norm_default_blocklist),
    torch.ops.aten.convolution.default: partial(guard_aten, aten_convolution_default_blocklist),
    torch.ops.aten.argmax.default: partial(guard_aten, aten_argmax_default_blocklist),
}



import os
import sys
import subprocess
from setuptools import setup, Extension, find_namespace_packages
from setuptools.command.build_ext import build_ext
from utils.get_torch_abi_flags import get_torch_abi_related_compiler_flags
import sysconfig
import torch


class CMakeExtension(Extension):
    def __init__(self, name, source_dir=".", cmake_args=None, **kwargs):
        Extension.__init__(self, name, sources=[])
        self.source_dir = os.path.abspath(source_dir)
        self.cmake_args = cmake_args or []


class CMakeBuild(build_ext):
    def build_extension(self, ext):
        if not isinstance(ext, CMakeExtension):
            super().build_extension(ext)
            return

        build_dir = os.path.join(self.build_temp, ext.name)
        os.makedirs(build_dir, exist_ok=True)

        # Configure CMake
        cmake_args = [
            f"-DCMAKE_BUILD_TYPE=Release",
            f"-DTORCH_INSTALL_PREFIX={sysconfig.get_paths()['purelib']}",
            f"-DCMAKE_PREFIX_PATH={torch.utils.cmake_prefix_path}",
            f"-DPYTHON_EXECUTABLE={sys.executable}",
            f"-DENABLE_LOCAL_TT_METAL_BUILD=ON",
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))}",
            f"-DOUTPUT_NAME={os.path.basename(self.get_ext_fullpath(ext.name))}",
            f"-G",
            f"Ninja",
        ]
        extra_cmake_flags = os.environ.get("CMAKE_FLAGS", "")
        extra_cmake_flags = extra_cmake_flags.split(";")
        if "-DENABLE_SUBMODULE_TT_METAL_BUILD=ON" in extra_cmake_flags:
            extra_cmake_flags.append("-DENABLE_LOCAL_TT_METAL_BUILD=OFF")

        torch_cxx_flags = get_torch_abi_related_compiler_flags()
        if torch_cxx_flags:
            flags_str = " ".join(torch_cxx_flags)
            extra_cmake_flags.append(f"-DCMAKE_CXX_FLAGS={flags_str}")

        if extra_cmake_flags:
            cmake_args.extend(extra_cmake_flags)

        cmake_args.extend(ext.cmake_args)

        # Build the extension
        subprocess.check_call(["cmake", ext.source_dir] + cmake_args, cwd=build_dir)
        subprocess.check_call(["cmake", "--build", ".", "--parallel"], cwd=build_dir)

        # Copy the extension to the correct location
        ext_path = self.get_ext_fullpath(ext.name)
        os.makedirs(os.path.dirname(ext_path), exist_ok=True)


setup(
    name="torch_ttnn_cpp_extension",
    version="0.1.0",
    packages=find_namespace_packages(include=["torch_ttnn*"]),
    ext_modules=[
        CMakeExtension(
            name="ttnn_device_extension",
            source_dir=".",
            cmake_args=[],
        ),
    ],
    cmdclass={"build_ext": CMakeBuild},
    python_requires=">=3.8",
)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
# The interface of this backend

import sys
import platform

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"

from .backend import ttnn_backend as backend
from .backend import TorchTtnnOption
import torch as _torch

_torch._dynamo.backends.registry.register_backend(name="ttnn", compiler_fn=backend)

# To wrap the ttnn ops
from .passes.lowering import target_wrappers


def _check_ttnn_requirements():
    """Check if the environment meets ttnn requirements."""
    if sys.platform != "linux":
        raise ImportError("ttnn is only supported on Linux. " "Current platform: {}".format(platform.platform()))

    python_version = sys.version_info
    if python_version.major != 3 or python_version.minor != 10:
        raise ImportError(
            "ttnn requires Python 3.10. "
            "Current Python version: {}.{}".format(python_version.major, python_version.minor)
        )


try:
    _check_ttnn_requirements()
    import ttnn
except ImportError as e:
    if "ttnn" not in str(e):
        raise
    print(
        "\nError: ttnn is not installed or not compatible with your system.\n"
        "To install ttnn, run:\n"
        "  pip install torch-ttnn[ttnn]\n\n"
        "Note: ttnn is only supported on Linux with Python 3.10.\n"
        "Current system: {}\n"
        "Current Python: {}.{}\n".format(platform.platform(), sys.version_info.major, sys.version_info.minor)
    )
    raise



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
import ttnn
from torch.fx.passes.infra.pass_base import PassBase, PassResult
from torch.fx.node import Node, map_arg
from .lowering import target_wrappers
import operator

from typing import Dict, List


@torch.fx.wrap
def deallocate(tensor):
    if isinstance(tensor, ttnn.Tensor):
        tensor.deallocate()
    return None


class TrackUnusedValues:
    """
    Adapted from CodeGen._gen_python_code in pytorch/torch/fx/graph.py
    Given a graph, tracks the last uses of each node in order to
    deallocate them appropriately. This is important for TTNN graphs because
    of limited L1 memory.

    Methods
    -------
    get_nodes_to_delete(user=""):
        Given a user, returns a list of nodes to deallocate

    Usage
    unused_values = TrackUnusedValues(graph)
    for node in graph.nodes:
        to_delete_nodes = unused_values.get_nodes_to_delete(node)
    """

    def __init__(self, graph):
        node_to_last_use: Dict[Node, Node] = {}
        self._user_to_last_uses: Dict[Node, List[Node]] = {}

        def register_last_uses(n: Node, user: Node):
            if n not in node_to_last_use:
                node_to_last_use[n] = user
                self._user_to_last_uses.setdefault(user, []).append(n)

        for node in reversed(graph.nodes):
            map_arg(node.args, lambda n: register_last_uses(n, node))
            map_arg(node.kwargs, lambda n: register_last_uses(n, node))

    def get_nodes_to_delete(self, user: Node):
        """
        Delete values after their last use. This ensures that values that are
        not used in the remainder of the code are freed and the memory usage
        of the code is optimal. Returns a string of all the values to delete.
        Returns an empty string if none.
        """
        if user.op == "placeholder":
            return None
        if user.op == "output":
            return None
        nodes_to_delete = self._user_to_last_uses.get(user, [])

        if len(user.users.keys()) == 0:
            # This node is not used by any others. however it's also not
            # removed by DCE since side-effect. We want to free it's outputs
            # right after its execution done to save memory.
            nodes_to_delete.append(user)

        if len(nodes_to_delete):
            return nodes_to_delete
        else:
            return None


class DeallocationPass(PassBase):
    def __init__(self):
        super().__init__()
        torch.fx.node.has_side_effect(deallocate)

    def call(self, gm: torch.fx.GraphModule):
        graph = gm.graph

        unused_values = TrackUnusedValues(graph)

        modified = False

        # Because we are inserting new nodes after the target node, we cannot iterate
        # through the graph nodes directly. Otherwise, we will get an infinite loop.
        # Instead we create a new list and iterate through that.
        node_list = list(graph.nodes)
        for node in node_list:
            if (to_delete_nodes := unused_values.get_nodes_to_delete(node)) is not None:
                for n in to_delete_nodes:
                    # Skip nodes that are references to other nodes
                    # We don't want to delete these too early
                    if node.target in [target_wrappers.pack_to_tuple, operator.getitem]:
                        continue
                    # Skip nodes that are cached between runs
                    elif n.meta.get("is_cached", False):
                        continue
                    with graph.inserting_after(node):
                        new_node = graph.call_function(deallocate, args=(n,))
                        modified = True

        if modified:
            # Becareful about dead code elimination at this point
            gm.graph.lint()
            gm.recompile()

        return PassResult(gm, True)



import torch
from torch.overrides import TorchFunctionMode
import logging

import ttnn_device_extension as ttnn_module

logging.info("Using pre-built ttnn_device_extension")

torch.utils.rename_privateuse1_backend("ttnn")


# The user will globally enable the below mode when calling this API
def enable_ttnn_device():
    m = TtnnDeviceMode()
    m.__enter__()
    # If you want the mode to never be disabled, then this function shouldn't return anything.
    return m


# This is a simple TorchFunctionMode class that:
# (a) Intercepts all torch.* calls
# (b) Checks for kwargs of the form `device="ttnn:i"`
# (c) Turns those into custom device objects: `device=ttnn_module.custom_device(i)`
# (d) Forwards the call along into pytorch.
class TtnnDeviceMode(TorchFunctionMode):
    def __torch_function__(self, func, types, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}
        if "device" in kwargs and "ttnn" in kwargs["device"]:
            device_and_idx = kwargs["device"].split(":")
            if len(device_and_idx) == 1:
                # Case 1: No index specified
                kwargs["device"] = ttnn_module.open_torch_device()
            else:
                # Case 2: The user specified a device index.
                device_idx = int(device_and_idx[1])
                kwargs["device"] = ttnn_module.open_torch_device(device_idx)
        with torch._C.DisableTorchFunction():
            return func(*args, **kwargs)



import torch
import re


def get_torch_abi_related_compiler_flags():
    """Extract compiler flags relaed to abi from torch.__config__.show()"""
    config_str = torch.__config__.show()

    # Extract C++ flags
    cxx_flags = []

    # Look for ABI flag, critical for compatibility
    abi_match = re.search(r"-D_GLIBCXX_USE_CXX11_ABI=(\d)", config_str)
    if abi_match:
        cxx_flags.append(f"-D_GLIBCXX_USE_CXX11_ABI={abi_match.group(1)}")

    fabi_match = re.search(r"-fabi-version=(\d+)", config_str)
    if fabi_match:
        cxx_flags.append(f"-fabi-version={fabi_match.group(1)}")

    return cxx_flags


if __name__ == "__main__":
    flags = get_torch_abi_related_compiler_flags()
    for flag in flags:
        print(flag, end=" ")



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
from torch.fx.passes.infra.pass_base import PassBase, PassResult
from enum import Enum


class ModelType(Enum):
    """Enumeration of model types differentiating between inference and training, forward and backward.

    :param INFERENCE: Model with this tag is for the forward pass of inference.
    :param TRAIN_FORWARD: Model wih this tag is for the forward pass of a training run.
    :param TRAIN_BACKWARD: Model with this tag is for the backward pass of a training run.
    """

    INFERENCE = 1
    TRAIN_FORWARD = 2
    TRAIN_BACKWARD = 3


def is_train_backward(gm):
    node_list = list(gm.graph.nodes)
    for node in node_list:
        if node.op == "call_function" and "backward" in str(node.target):
            return True

    return False


def is_train_forward(gm):
    # Assume training forward calls are differentiated by directly returning one of the inputs to be used for calculating gradients
    # If this assumption fails in the future, we will need to update this function
    outputs = [node for node in gm.graph.nodes if node.op == "output"]
    for node in outputs:
        placeholder_args = [arg for arg in node.args[0] if arg is not None and arg.op == "placeholder"]
        if len(placeholder_args) > 0:
            return True

    return False


class GraphModuleAnalysisPass(PassBase):
    def __init__(self):
        super().__init__()

    def call(self, gm: torch.fx.GraphModule):
        """Marks the GraphModule as either training forward, training backward, or inference (forward).

        This relies on commonalities between training forward, backward, and inference graphs. Namely, backward passes call backward versions of the forward functions to calculate gradients. Training forward passes return inputs unchanged. Inference forward functions do neither of these. It would be cleaner if we could just use something like `torch.is_grad_enabled()` or `gm.training` instead, but these appear to be inaccurate by the time the GraphModule is passed to our backend.

        :param gm: Graph module for the function being compiled.
        :return: Pass result with the updated graph module with metadata indicating the type of graph being compiled.
        :rtype: PassResult[torch.fx.GraphModule, bool]
        """

        modified = False

        # check nodes for backward function call
        if is_train_backward(gm):
            gm.meta["graph_type"] = ModelType.TRAIN_BACKWARD
        elif is_train_forward(gm):
            gm.meta["graph_type"] = ModelType.TRAIN_FORWARD
        else:
            gm.meta["graph_type"] = ModelType.INFERENCE

        return PassResult(gm, modified)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
import ttnn
import math
from torch._guards import detect_fake_mode
from torch._subclasses.fake_tensor import unset_fake_temporarily

from torch_ttnn.utils import (
    GraphCleanup,
    TtnnBfloat16,
    TtnnUint32,
    TtnnInt32,
    TtnnUint32,
    TtnnDevice,
    TtnnL1MemoryConfig,
    TtnnRowMajorLayout,
    TtnnTileLayout,
    get_emplace_custom_object_in_graph,
    get_shape,
    get_dtype,
    get_arg,
    get_meta_val_attr,
)
import numpy as np
import torch_ttnn.metrics as metrics

from torch.fx.passes.infra.pass_base import PassBase, PassResult
import torch.fx.traceback as fx_traceback
from . import target_wrappers
from .to_tt_guard import can_lowering_to_ttnn
import operator

relational_scalar_ops = {
    torch.ops.aten.eq.Scalar: ttnn.eq,
    torch.ops.aten.lt.Scalar: ttnn.lt,
    torch.ops.aten.le.Scalar: ttnn.le,
    torch.ops.aten.gt.Scalar: ttnn.gt,
    torch.ops.aten.ge.Scalar: ttnn.ge,
    torch.ops.aten.ne.Scalar: ttnn.ne,
}

# Workaround: If an arg of the model output is argmax then skip conversion
# TODO(kevinwuTT): Handle this case with ttnn ops
int_output_ops = [
    torch.ops.aten.argmax.default,
    torch.ops.aten.argmin.default,
]


ops_incompatible_with_grayskull = {
    torch.ops.aten.ceil.default,
    torch.ops.aten.floor.default,
    torch.ops.aten.round.default,
    torch.ops.aten.round.decimals,
    torch.ops.aten.trunc.default,
}


def is_target_incompatible_with_grayskull(target, device):
    return ttnn.device.is_grayskull(device) and target in ops_incompatible_with_grayskull


def are_args_from_int_output_ops(args):
    for arg in args:
        if isinstance(arg, torch.fx.proxy.Proxy):
            if arg.node.target in int_output_ops:
                return True


def create_call_function(transformer, target, args, kwargs):
    transformer.call_function(target, args, kwargs)


def map_args_to_kwargs(args, kw_map, default_none=False):
    """
    Args:
        args (List): the input arguments
        kw_map (List[str]): the list of ("index in args" : int, "keyword name" : str)
        default_none (bool): put none if the value is missing in args; otherwise don't add the key
    Returns:
        Dict[str, Any]: the extracted values
    """
    kwargs = {}
    for idx, kw in kw_map:
        if idx >= len(args) and not default_none:
            break
        kwargs[kw] = args[idx] if idx < len(args) else None
    return kwargs


def is_getitem_0_only_user(node):
    return all(
        user.op == "call_function" and user.target.__name__ == "getitem" and user.args[1] == 0
        for user in node.users.keys()
    )


# Convert (n, c, x, ...) to (n, x, ..., c)
def insert_ncx_to_nxc(g, input_tensor):
    input_shape = input_tensor.meta["val"].size()
    target_permute = [0] + list(range(2, len(input_shape))) + [1]
    return g.call_function(ttnn.permute, (input_tensor, target_permute))


# Convert sharded (1, 1, nx..., c) to interleaved (n, c, x, ...)
def insert_sharded_nxc_to_ncx(g, output_tensor, output_shape):
    output_tensor = g.call_function(ttnn.sharded_to_interleaved, (output_tensor, TtnnL1MemoryConfig()))
    target_shape = (output_shape[0], *output_shape[2:], output_shape[1])
    output_tensor = g.call_function(ttnn.reshape, (output_tensor, target_shape))
    target_permute = [0, len(output_shape) - 1] + list(range(1, len(output_shape) - 1))
    return g.call_function(ttnn.permute, (output_tensor, target_permute))


TTNN_POINTWISE_UNARY_OPS = {
    torch.ops.aten.abs.default: ttnn.abs,
    torch.ops.aten.acos.default: ttnn.acos,
    torch.ops.aten.acosh.default: ttnn.acosh,
    torch.ops.aten.asin.default: ttnn.asin,
    torch.ops.aten.asinh.default: ttnn.asinh,
    torch.ops.aten.atan.default: ttnn.atan,
    torch.ops.aten.atanh.default: ttnn.atanh,
    torch.ops.aten.ceil.default: ttnn.ceil,
    torch.ops.aten.clamp.default: ttnn.clip,
    torch.ops.aten.cos.default: ttnn.cos,
    torch.ops.aten.cosh.default: ttnn.cosh,
    torch.ops.aten.elu.default: ttnn.elu,
    torch.ops.aten.erf.default: ttnn.erf,
    torch.ops.aten.exp.default: ttnn.exp,
    torch.ops.aten.expm1.default: ttnn.expm1,
    torch.ops.aten.floor.default: ttnn.floor,
    torch.ops.aten.gelu.default: ttnn.gelu,
    torch.ops.aten.hardsigmoid.default: ttnn.hardsigmoid,
    torch.ops.aten.hardswish.default: ttnn.hardswish,
    torch.ops.aten.isinf.default: ttnn.isinf,
    torch.ops.aten.isnan.default: ttnn.isnan,
    torch.ops.aten.log.default: ttnn.log,
    torch.ops.aten.log10.default: ttnn.log10,
    torch.ops.aten.log1p.default: ttnn.log1p,
    torch.ops.aten.log2.default: ttnn.log2,
    torch.ops.aten.logical_not.default: ttnn.logical_not,
    torch.ops.aten.neg.default: ttnn.neg,
    torch.ops.aten.reciprocal.default: ttnn.reciprocal,
    torch.ops.aten.remainder.Scalar: ttnn.remainder,
    torch.ops.aten.relu.default: ttnn.relu,
    torch.ops.aten.round.decimals: ttnn.round,
    torch.ops.aten.rsqrt.default: ttnn.rsqrt,
    torch.ops.aten.sigmoid.default: ttnn.sigmoid,
    torch.ops.aten.sign.default: ttnn.sign,
    torch.ops.aten.sin.default: ttnn.sin,
    torch.ops.aten.sinh.default: ttnn.sinh,
    torch.ops.aten.silu.default: ttnn.silu,
    torch.ops.aten.sqrt.default: ttnn.sqrt,
    torch.ops.aten.tan.default: ttnn.tan,
    torch.ops.aten.tanh.default: ttnn.tanh,
    torch.ops.aten.tril.default: ttnn.tril,
    torch.ops.aten.trunc.default: ttnn.trunc,
}


class ReplaceMoreTt(torch.fx.Transformer):
    def __init__(self, module, device, use_less_ttnn_op_types):
        super().__init__(module)
        get_meta_for_op = lambda op: {node.name: node.meta for node in self.module.graph.nodes if node.op == op}
        self._input_node_meta = get_meta_for_op("placeholder")
        self._get_attr_meta = get_meta_for_op("get_attr")
        self.device = device
        self.use_less_ttnn_op_types = use_less_ttnn_op_types

    def transform(self):
        old_meta = self.module.meta
        result = super().transform()
        result.meta = old_meta
        return result

    def get_attr(self, target, args, kwargs):
        # Restore original metadata for get_attr nodes
        proxy = super().get_attr(target, args, kwargs)
        if proxy.node.name in self._get_attr_meta:
            proxy.node.meta = self._get_attr_meta[proxy.node.name]
        return proxy

    def placeholder(self, target, args, kwargs):
        # Restore original metadata for placeholder nodes
        proxy = super().placeholder(target, args, kwargs)
        if proxy.node.name in self._input_node_meta:
            proxy.node.meta = self._input_node_meta[proxy.node.name]
        return proxy

    def call_function_prop_meta(self, target, args=(), kwargs={}):
        call_func = super().call_function(target, args, kwargs)
        meta = fx_traceback.get_current_meta()
        if meta is not None:
            call_func.node.meta = meta
            if hasattr(self.old_target, "_schema"):
                call_func.node.meta["original_input_variations"] = metrics.collect_input_variation(
                    self.old_target, self.old_args, self.old_kwargs
                )

        return call_func

    def call_function(self, target, args, kwargs):
        # FIXME: `call_function_prop_meta` should be moved to an inner function here instead of using member variables
        self.old_target = target
        self.old_args = args
        self.old_kwargs = kwargs

        class PseudoNode:
            def __init__(self, target, args, kwargs):
                self.target = target
                self.args = args
                self.kwargs = kwargs

        pseudo_node = PseudoNode(target, args, kwargs)
        if not can_lowering_to_ttnn(pseudo_node):
            return self.call_function_prop_meta(target, args, kwargs)

        if are_args_from_int_output_ops(args) or is_target_incompatible_with_grayskull(target, self.device):
            return self.call_function_prop_meta(target, args, kwargs)

        ############################################################
        # Matrix multiplication
        ############################################################
        if target == torch.ops.aten.addmm.default:
            # TODO(kevinwuMCW): include beta, alpha, and optional args
            mm = self.call_function_prop_meta(ttnn.matmul, (args[1], args[2]), kwargs)
            return self.call_function_prop_meta(ttnn.add, (args[0], mm), kwargs)

        if target == torch.ops.aten.bmm.default:
            return self.call_function_prop_meta(ttnn.matmul, args, kwargs)

        if target == torch.ops.aten.linear.default:
            return self.call_function_prop_meta(ttnn.linear, args, kwargs)

        if target == torch.ops.aten.mm.default:
            return self.call_function_prop_meta(ttnn.matmul, args, kwargs)

        ############################################################
        # Pointwise unary
        ############################################################
        if target == torch.ops.aten.hardtanh.default:
            # aten.hardtanh args are positional but ttnn.clip uses kw args
            new_kwargs = map_args_to_kwargs(args, ((1, "min_val"), (2, "max_val")), default_none=True)
            new_args = (args[0],)
            return self.call_function_prop_meta(ttnn.hardtanh, new_args, new_kwargs)

        if target == torch.ops.aten._softmax.default:
            kwargs = {
                "numeric_stable": True,
                **kwargs,
            }
            return self.call_function_prop_meta(ttnn.softmax, args[:2], kwargs)

        ############################################################
        # Pointwise binary
        ############################################################
        if target in relational_scalar_ops:
            return self.call_function_prop_meta(relational_scalar_ops[target], args, kwargs)

        if target == torch.ops.aten.atan2.default:
            return self.call_function_prop_meta(ttnn.atan2, args, kwargs)

        if target == torch.ops.aten.leaky_relu.default:
            if len(args) < 2:
                args = (args[0], 0.01)
            return self.call_function_prop_meta(ttnn.leaky_relu, args, kwargs)

        if target == torch.ops.aten.maximum.default:
            if get_shape(None, args[0]) != get_shape(None, args[1]):
                # see tt-metal#12852
                return self.call_function_prop_meta(target, args, kwargs)
            return self.call_function_prop_meta(ttnn.maximum, args, kwargs)

        if target == torch.ops.aten.minimum.default:
            if get_shape(None, args[0]) != get_shape(None, args[1]):
                # see tt-metal#12852
                return self.call_function_prop_meta(target, args, kwargs)
            return self.call_function_prop_meta(ttnn.minimum, args, kwargs)

        if target in [torch.ops.aten.pow.Scalar, torch.ops.aten.pow.Tensor_Scalar, torch.ops.aten.pow.Tensor_Tensor]:
            if target == torch.ops.aten.pow.Tensor_Tensor:
                shape0 = get_shape(None, args[0])
                shape1 = get_shape(None, args[1])
                if (shape0 != shape1) or shape0 is None or shape1 is None:
                    # not support broadcasting
                    return self.call_function_prop_meta(target, args, kwargs)
            return self.call_function_prop_meta(ttnn.pow, args, kwargs)

        if target == torch.ops.aten.rsub.Scalar:
            return self.call_function_prop_meta(ttnn.rsub, args, kwargs)

        if target == torch.ops.aten.rsub.Tensor:
            # TODO(kevinwuMCW): handle alpha parameter if exists
            return self.call_function_prop_meta(ttnn.sub, (args[1], args[0]), kwargs)

        if target == torch.ops.aten.xlogy.Tensor:
            return self.call_function_prop_meta(ttnn.xlogy, args, kwargs)

        ############################################################
        # Pointwise ternary
        ############################################################
        if target == torch.ops.aten.addcdiv.default:
            value = kwargs.pop("value", 1.0)
            return self.call_function_prop_meta(ttnn.addcdiv, args + (value,), kwargs)

        if target == torch.ops.aten.addcmul.default:
            value = kwargs.pop("value", 1.0)
            return self.call_function_prop_meta(ttnn.addcmul, args + (value,), kwargs)

        if target == torch.ops.aten.where.self:
            return self.call_function_prop_meta(ttnn.where, args, kwargs)

        ############################################################
        # Reduction
        ############################################################
        if target == torch.ops.aten.mean.dim:
            return self.call_function_prop_meta(ttnn.mean, args, kwargs)

        if target == torch.ops.aten.min.default:
            return self.call_function_prop_meta(ttnn.min, args, kwargs)

        ############################################################
        # Other ops
        ############################################################
        if target == torch.ops.aten._adaptive_avg_pool2d.default:
            arg0_size = None
            if hasattr(args[0], "node") and hasattr(args[0].node, "meta"):
                arg0_size = list(args[0].node.meta["val"].size())
            output_size = None
            if len(args) > 1:
                output_size = args[1]
            elif "output_size" in kwargs:
                output_size = kwargs["output_size"]
            if arg0_size is not None and output_size is not None and arg0_size[-2:] == output_size:
                return args[0]
            # TODO: no ttnn op can convert _adaptive_avg_pool2d
            return self.call_function_prop_meta(target, args, kwargs)

        if target == torch.ops.aten.detach.default:
            return args[0]

        if target == torch.ops.aten.as_strided.default:
            unpack = lambda tensor, size, stride, offset=0: (tensor, size, stride, offset)
            tensor, size, stride, offset = unpack(*args)
            ndims = len(size)

            if offset != 0 or len(stride) != ndims:
                return self.call_function_prop_meta(target, args, kwargs)

            dummy = torch.empty(0, *size, dtype=torch.bool)

            for i in range(ndims):
                if stride[i] != dummy.stride(i + 1) and size[i] != 1:
                    return self.call_function_prop_meta(target, args, kwargs)

            return self.call_function_prop_meta(ttnn.reshape, (tensor, size))

        if target == torch.ops.aten.permute.default:
            return self.call_function_prop_meta(ttnn.permute, args, kwargs)

        return self.call_function_prop_meta(target, args, kwargs)


def torch_dtype_to_ttnn_dtype(dtype: torch.dtype):
    # Add newly supported dtypes here:
    dtype_map = {
        torch.float32: TtnnBfloat16(),  # Should this be changed to TtnnFloat32?
        torch.bfloat16: TtnnBfloat16(),
    }
    if dtype in dtype_map:
        return dtype_map.get(dtype)
    else:
        raise RuntimeError(f"Missing conversion from torch.dtype: {dtype} to Ttnn dtype.")


# override some functions from torch.fx.graph.Graph
class GraphWrapper:
    def __init__(self, gm, node):
        self.gm = gm
        self.g = node.graph
        self.node = node

    def call_function(self, target, args=(), kwargs={}):
        new_node = self.g.call_function(target, args, kwargs)
        new_node.meta = dict(self.node.meta)
        if hasattr(self.node.target, "_schema"):
            new_node.meta["original_input_variations"] = metrics.collect_input_variation_from_node(self.node)
        new_node.meta["val"] = self._get_output_val(new_node)
        return new_node

    def inserting_before(self, node):
        return self.g.inserting_before(node)

    def _get_output_val(self, node):
        args = node.args
        # Compute on fake tensor to derive the output fake tensor
        if node.target == ttnn.add:
            return torch.add(self._get_val(args[0]), self._get_val(args[1]))
        if node.target == ttnn.mul:
            return torch.mul(self._get_val(args[0]), self._get_val(args[1]))
        if node.target == ttnn.sub:
            return torch.sub(self._get_val(args[0]), self._get_val(args[1]))
        if node.target in [ttnn.rsqrt, ttnn.sharded_to_interleaved]:
            return self._get_val(args[0])
        # When accessing to self.node.meta, we assume the GraphWrapper is created on a pytorch op mapped to that ttnn op,
        # so we can use the output shape from it
        if node.target in [ttnn.max_pool2d, ttnn.avg_pool2d, target_wrappers.conv]:
            output_val = self._get_val(self.node)
            output_tensor = output_val[0] if isinstance(output_val, tuple) else output_val
            output_shape = list(output_tensor.size())
            return output_tensor.new_empty((1, 1, output_shape[0] * math.prod(output_shape[2:]), output_shape[1]))
        if node.target == ttnn.reshape:
            return torch.reshape(self._get_val(args[0]), args[1])
        if node.target == ttnn.permute:
            return torch.permute(self._get_val(args[0]), args[1])
        return self._get_val(node)

    def _get_val(self, obj):
        if not isinstance(obj, torch.fx.node.Node):
            return obj
        if obj.op == "get_attr":
            val = getattr(self.gm, obj.target)
            fake_mode = detect_fake_mode()
            if isinstance(val, torch.Tensor) and fake_mode is not None:
                val = fake_mode.fake_tensor_converter.from_real_tensor(fake_mode, val)
            return val
        return obj.meta["val"]


def ReplaceMoreTtManually(gm: torch.fx.GraphModule, device, use_less_ttnn_op_types: bool) -> torch.fx.GraphModule:
    nodes = list(gm.graph.nodes)
    for node in nodes:
        if not can_lowering_to_ttnn(node):
            continue
        g = GraphWrapper(gm, node)

        def rewrite_node(node):
            args = node.args
            kwargs = node.kwargs

            def batch_norm_post_process(output, shape, weight, bias):
                if weight is not None:
                    weight = g.call_function(ttnn.reshape, (weight, shape))
                    output = g.call_function(ttnn.mul, (output, weight))
                if bias is not None:
                    bias = g.call_function(ttnn.reshape, (bias, shape))
                    output = g.call_function(ttnn.add, (output, bias))
                return output

            # Non-training batch normalization
            def batch_norm_inference(input, weight, bias, mean, var, momentum, eps):
                assert is_getitem_0_only_user(node), "non-training batch_norm should only have first return value used"
                shape = input.meta["val"].size()
                shape = (1, shape[1]) + (1,) * (len(shape) - 2)
                invstd = g.call_function(ttnn.rsqrt, (g.call_function(ttnn.add, (var, eps)),))
                invstd = g.call_function(ttnn.reshape, (invstd, shape))
                mean = g.call_function(ttnn.reshape, (mean, shape))
                output = g.call_function(ttnn.sub, (input, mean))
                output = g.call_function(ttnn.mul, (output, invstd))
                output = batch_norm_post_process(output, shape, weight, bias)
                return g.call_function(target_wrappers.pack_to_tuple, (output,))

            def lower_binary_eltwise(fn, args):
                shapes = get_shape(gm, args[0]), get_shape(gm, args[1])

                if any(s is None for s in shapes):
                    return None

                if max(map(len, shapes)) > 4 and shapes[0][-3:-2] != shapes[1][-3:-2]:
                    return None

                def cast_bf16(target, other):
                    if get_dtype(target) in [torch.int32, torch.int64] and get_dtype(other) in [
                        torch.bfloat16,
                        torch.float32,
                    ]:
                        return g.call_function(ttnn.typecast, args=(target, TtnnBfloat16()))
                    else:
                        return target

                new_args = (cast_bf16(args[0], args[1]), cast_bf16(args[1], args[0]))

                return g.call_function(fn, new_args)

            if node.target == torch.ops.aten.add.Tensor:
                return lower_binary_eltwise(ttnn.add, args)

            if node.target == torch.ops.aten.mul.Tensor:
                return lower_binary_eltwise(ttnn.mul, args)

            if node.target == torch.ops.aten.sub.Tensor:
                return lower_binary_eltwise(ttnn.sub, args)

            if node.target == torch.ops.aten.tanh.default:
                # TODO: Remove once tanh accuracy implementation is realized as a device operation in tt-metal
                new_kwargs = kwargs.copy()
                new_kwargs["accuracy"] = True
                kwargs = new_kwargs

            if node.target in TTNN_POINTWISE_UNARY_OPS:
                return g.call_function(TTNN_POINTWISE_UNARY_OPS[node.target], args, kwargs)

            if node.target == torch.ops.aten.round.default:
                return g.call_function(ttnn.round, (args[0],), {"decimals": 0})

            if node.target == torch.ops.aten.clone.default:
                # Only convert if the input is from graph arguments and node is also returned as an output
                # Otherwise, optimize node out and return input
                node_users = list(node.users.keys())
                if not (args[0].op == "placeholder" and len(node_users) == 1 and node_users[0].op == "output"):
                    return args[0]

                arg_metadata = node.meta["val"]
                try:
                    ttnn_dtype = torch_dtype_to_ttnn_dtype(arg_metadata.dtype)
                except:
                    return None
                # Add additional logic to choose the appropriate memory_config type: DRAM or L1
                return g.call_function(target_wrappers.clone, args=(args[0],))

            if node.target == torch.ops.aten.native_layer_norm.default:
                input_tensor, normalized_shape, weight, bias, epsilon = args

                in_tensor_shape = input_tensor.meta["val"].size()
                mean_rstd_shape = node.meta["val"][1].size()
                torch_dtype = input_tensor.meta["val"].dtype
                ttnn_dtype = torch_dtype_to_ttnn_dtype(torch_dtype)
                norm_dims = len(normalized_shape)

                # aten.native_layer_norm keeps the normalized dimension for mean and rstd,
                # but ttnn.moreh.layer_norm does not.
                ttnn_mean_rstd_shape = in_tensor_shape[:-norm_dims]

                # At the time of implementation, ttnn.layer_norm is marginally faster
                # than moreh.layer_norm if only layer_norm output is needed.
                # Check if mean and/or rstd is used. The output from the wrapper
                # can be a tuple of 1, 2, or 3 items. The 1st item will always be layer_norm.
                # The 2nd can be mean or rstd. The 3rd will always be rstd if it exists.
                # Since `getitem` always follows layer_norm, the number of direct users of this
                # node should always be 1, 2, or 3.
                node_users = list(node.users.keys())
                use_mean = False
                use_rstd = False
                # check the 2nd item if it's mean or rstd
                if len(node_users) > 1:
                    getitem = node_users[1]
                    if getitem.args[1] == 1:
                        use_mean = True
                    if getitem.args[1] == 2:
                        use_rstd = True
                # if the 3rd item exists, it will always be rstd
                if len(node_users) > 2:
                    use_rstd = True

                new_node = g.call_function(
                    target_wrappers.native_layer_norm,
                    (
                        input_tensor,
                        in_tensor_shape,
                        mean_rstd_shape,
                        ttnn_mean_rstd_shape,
                        ttnn_dtype,
                        norm_dims,
                        weight,
                        bias,
                        epsilon,
                        use_mean,
                        use_rstd,
                        TtnnDevice(),
                    ),
                )

                # update metadata since original op does not have dtype for mean and rstd outputs
                for node_user in node_users:
                    node_user.meta["val"] = node_user.meta["val"].to(torch_dtype)

                return new_node

            if node.target == torch.ops.aten.zeros.default:
                return g.call_function(ttnn.zeros, args=args, kwargs={"device": TtnnDevice()})

            if node.target == torch.ops.aten.zeros_like.default:
                # TODO(#280): Doesn't support 1D output tensor in tile layout (#280)
                if len(node.meta["val"].shape) < 2:
                    return None
                return g.call_function(ttnn.zeros_like, args, kwargs={"device": TtnnDevice()})

            if node.target == torch.ops.aten.ones.default:
                return g.call_function(ttnn.ones, args=args, kwargs={"device": TtnnDevice()})

            """
            # NOTE(kevinwuTT): aten.arange.default starts with 0 which is unsupported by ttnn.arange at the moment
            if node.target == torch.ops.aten.arange.default:
                # start = 0, step = 1
                new_args = (0, args[0], 1)
                return g.call_function(ttnn.arange, args=new_args)
            """
            if node.target == torch.ops.aten.arange.start:
                # NOTE(kevinwuTT): ttnn.arange does not support starting values smaller than 2 currently
                if args[0] >= 2:
                    # step = 1
                    new_args = (args[0], args[1], 1)
                    return g.call_function(ttnn.arange, args=new_args)
                return None

            if node.target == torch.ops.aten.arange.start_step:
                # NOTE(kevinwuTT): ttnn.arange does not support starting values smaller than 2 currently
                if args[0] >= 2:
                    new_args = (args[0], args[1], args[2])
                    return g.call_function(ttnn.arange, args=new_args)
                return None

            if node.target == torch.ops.aten.eq.Tensor:
                # Combine this with relational_scalar_ops
                if np.prod(args[1].meta["val"].size()) != 1:
                    return g.call_function(
                        ttnn.eq,
                        args=args,
                        kwargs={},
                    )
                return None

            if node.target == torch.ops.aten.full.default:
                new_kwargs = {
                    "fill_value": args[1],
                    "device": TtnnDevice(),
                    "layout": TtnnTileLayout(),
                }
                return g.call_function(ttnn.full, args=(args[0],), kwargs=new_kwargs)

            if node.target == torch.ops.aten.baddbmm.default:
                # out = beta * input + alpha * (batch1 @ batch2)
                # if beta is 0, input is ignored, and nan and inf in it will not be propogated
                new_node = g.call_function(ttnn.matmul, args=(args[1], args[2]))
                alpha = kwargs.get("alpha", 1)
                beta = kwargs.get("beta", 1)

                if alpha != 1:
                    new_node = g.call_function(ttnn.multiply, args=(new_node, alpha))

                if beta == 0:
                    return new_node

                if beta == 1:
                    return g.call_function(ttnn.add, args=(args[0], new_node))

                beta_node = g.call_function(ttnn.multiply, args=(args[0], beta))
                return g.call_function(ttnn.add, args=(beta_node, new_node))

            if node.target == torch.ops.aten.embedding.default:
                tiled = args[1].meta.get("val")
                tiled = False if tiled is None else tiled.size()[-1] % ttnn.TILE_SIZE == 0
                layout = TtnnTileLayout() if tiled else TtnnRowMajorLayout()
                tensor = g.call_function(ttnn.embedding, (args[1], args[0]), {"layout": layout})
                return tensor if tiled else g.call_function(ttnn.to_layout, (tensor, TtnnTileLayout()))

            if node.target == torch.ops.aten.div.Tensor:
                if isinstance(args[1], (float, int)):
                    return g.call_function(ttnn.mul, (args[0], 1.0 / args[1]), {})

                if get_shape(gm, args[0]) == get_shape(gm, args[1]):
                    return g.call_function(ttnn.div, args, {})

                recip = g.call_function(ttnn.reciprocal, (args[1],), {})
                return g.call_function(ttnn.mul, (args[0], recip), {})

            if node.target == torch.ops.aten.floor_divide.default:
                return g.call_function(ttnn.floor_div, args, {})

            if node.target == torch.ops.aten.expand.default:
                if not (hasattr(args[0], "meta") and "val" in args[0].meta and hasattr(args[0].meta["val"], "size")):
                    return None
                if not (hasattr(node, "meta") and "val" in node.meta and hasattr(node.meta["val"], "size")):
                    return None
                input_tensor_shape = args[0].meta["val"].size()
                output_shape = node.meta["val"].size()
                if input_tensor_shape.numel() == output_shape.numel():
                    if input_tensor_shape != output_shape:
                        return g.call_function(ttnn.reshape, args=(args[0], list(output_shape)))
                    else:
                        return args[0]

                input_shape = np.ones(len(output_shape), dtype=int)
                input_shape[-len(input_tensor_shape) :] = input_tensor_shape
                multiplier = np.array(output_shape) // np.array(input_shape)

                np_output_shape = np.array(list(output_shape))
                expand_multiplier = np_output_shape[np_output_shape > 1] // input_shape[np_output_shape > 1]
                expand_index = np.where(expand_multiplier > 1)[0]

                if input_tensor_shape[-1] % 2 == 0 and np.all(expand_index == np.arange(len(expand_index))):
                    return g.call_function(ttnn.expand, args=(args[0], list(output_shape)))

                # aten.expand and ttnn.repeat has different meaning for their `shape` argument
                # aten.expand: the desired output shape, where respective singleton dims are broadcasted
                # ttnn.repeat: the number of times to repeat a respective singleton dim
                if len(input_tensor_shape) == len(output_shape):
                    return g.call_function(target_wrappers.repeat, args=(args[0], multiplier.tolist()))

                return None

            if node.target == torch.ops.aten.slice.Tensor:
                tensor, dim, start, end, *step = args
                if tensor.op == "get_attr":
                    value = getattr(gm, tensor.target)
                    input_size = list(value.size())
                else:
                    input_size = list(tensor.meta["val"].size())

                [step] = step or [1]
                rank = len(input_size)

                if dim >= rank:
                    return None

                # Check if no-op, just return the input tensor
                if start == 0 and end >= input_size[dim] and step == 1:
                    return tensor

                slice_start = np.zeros(rank, dtype=int)
                slice_end = input_size

                # For negative end values
                if end < 0:
                    end = input_size[dim] + end
                else:
                    end = np.clip(end, a_min=None, a_max=input_size[dim])

                slice_start[dim] = start
                slice_end[dim] = end

                if step != 1:
                    array_step = [1] * len(input_size)
                    array_step[dim] = step

                kwargs = {
                    "slice_step": array_step if step != 1 else None,
                }

                return g.call_function(ttnn.slice, (tensor, [*slice_start], [*slice_end]), kwargs)

            if node.target == torch.ops.aten.repeat.default:
                tensor, sizes = args
                shape = get_shape(gm, tensor)
                if shape is None:
                    return None

                if np.prod(sizes) == 1:
                    return tensor

                # Current repeat implementation requires aligned last dim when repeating on last dim
                if sizes[-1] == 1 or shape[-1] % 16 == 0:
                    return g.call_function(target_wrappers.repeat, args)

                return None

            if node.target == torch.ops.aten.squeeze.dim or node.target == torch.ops.aten.squeeze.default:
                return g.call_function(ttnn.squeeze, args=args, kwargs=kwargs)

            if node.target == torch.ops.aten.unsqueeze.default:
                output_shape_num_element = node.meta["val"].numel()
                if output_shape_num_element == 0:
                    return args[0]
                output_size = list(node.meta["val"].size())
                return g.call_function(ttnn.reshape, args=(args[0], output_size))

            if node.target in [torch.ops.aten.transpose.int, torch.ops.aten.t.default]:
                rank = len(node.meta["val"].size())
                if rank < 2:
                    # Less 2D transpose is no-op
                    return args[0]
                if node.target == torch.ops.aten.t.default:
                    assert rank == 2, "Input tensor should only be 2D here"
                    dim0 = 0
                    dim1 = 1
                else:
                    dim0 = args[1]
                    dim1 = args[2]

                dim0 = (dim0 + rank) % rank
                dim1 = (dim1 + rank) % rank

                # directly transpose only support rank <= 4
                if rank <= 4:
                    return g.call_function(ttnn.transpose, args=(args[0], dim0, dim1))

                # Try to reshape. Find the new shape here
                #   eg. the original shape [A, B, C, D, E] shape[dim0] = A, shape[dim1] = D
                #   reshape to [A, (B*C), D, E]
                original_shape = list(args[0].meta["val"].size())
                dim0, dim1 = (dim0, dim1) if dim0 < dim1 else (dim1, dim0)

                new_dim0, new_dim1 = 0, 1
                new_shape = []
                if len(original_shape[:dim0]) > 0:
                    new_shape.append(np.prod(original_shape[:dim0]))
                    new_dim0 += 1
                    new_dim1 += 1

                new_shape.append(original_shape[dim0])

                if len(original_shape[dim0 + 1 : dim1]) > 0:
                    new_shape.append(np.prod(original_shape[dim0 + 1 : dim1]))
                    new_dim1 += 1

                new_shape.append(original_shape[dim1])

                if dim1 + 1 < rank:
                    new_shape.append(np.prod(original_shape[dim1 + 1 :]))

                # if the rank of reshaped result still <= 4, then do
                #    reshape->transpose->reshape
                if len(new_shape) <= 4:
                    reshaped = g.call_function(ttnn.reshape, args=(args[0], list(new_shape)))
                    transposed = g.call_function(ttnn.transpose, args=(reshaped, new_dim0, new_dim1))
                    output_shape = list(node.meta["val"].size())
                    return g.call_function(ttnn.reshape, args=(transposed, list(output_shape)))

                return None

            if node.target == torch.ops.aten.constant_pad_nd.default:
                input, pad = args[0], args[1]
                if len(args) > 2:
                    value = args[2]
                else:
                    value = 0
                # TODO(#516)
                if any(p < 0 for p in pad):
                    return None
                input_shape = input.meta["val"].size()
                output_shape = node.meta["val"].size()
                rank = len(input_shape)
                full_pad = [(0, 0)] * (rank - len(pad))
                # The order of pad from pytorch is reversed
                full_pad += [(pad[i], pad[i + 1]) for i in range(0, len(pad), 2)][::-1]

                # TODO(#514)
                if rank > 4:
                    return None

                # Change layout to row-major for non-tile-size-aligned tensor or front padding
                if (
                    rank < 2
                    or input_shape[-1] % ttnn.TILE_SIZE != 0
                    or input_shape[-2] % ttnn.TILE_SIZE != 0
                    or full_pad[-1][1] % ttnn.TILE_SIZE != 0
                    or full_pad[-2][1] % ttnn.TILE_SIZE != 0
                ) or not all(f == 0 for f, _ in full_pad):
                    input = g.call_function(ttnn.to_layout, args=(input, TtnnRowMajorLayout()))

                return g.call_function(ttnn.pad, args=(input, full_pad, value))

            if node.target in [torch.ops.aten.view.default, torch.ops.aten._unsafe_view.default]:
                input_tensor_num_element = args[0].meta["val"].numel()
                output_shape_num_element = node.meta["val"].numel()
                if input_tensor_num_element == 0 or output_shape_num_element == 0:
                    return None

                input_shape = args[0].meta["val"].shape
                output_shape = args[1]

                # To lower as ttnn.experimental.view:
                # * the last dimension must not change
                # * In Layout::TILE the second last two dimensions must not change OR there is no padding on the second last dimension
                # Be cautious and assume Layout::TILE for now

                last_dimension_constant = input_shape[-1] == output_shape[-1]
                second_last_dimension_constant = (
                    len(input_shape) > 1 and len(output_shape) > 1 and input_shape[-2] == output_shape[-2]
                )
                second_last_dimension_unpadded = (
                    len(input_shape) > 1
                    and len(output_shape) > 1
                    and input_shape[-2] % ttnn.TILE_SIZE == 0
                    and output_shape[-2] % ttnn.TILE_SIZE == 0
                )

                if last_dimension_constant and (second_last_dimension_constant or second_last_dimension_unpadded):
                    return g.call_function(ttnn.experimental.view, (args[0], args[1]), {})
                else:
                    return g.call_function(ttnn.reshape, (args[0], args[1]), {})

            if node.target == torch.ops.aten.split.Tensor:
                if len(args[0].meta["val"].size()) == 1:
                    # For example, the input shape original is [768]
                    # But due to issue #390 it become [1, 768] and cause failed
                    # remove this part once #390 is solved
                    return None

                # convert relative split dim to absolute
                dim = args[2] if len(args) > 2 else 0
                if dim >= 0:
                    split_dim = dim
                else:
                    split_dim = len(args[0].meta["val"].size()) + dim

                # convert from PyTorch size of chunk to ttnn number of chunks
                if isinstance(args[1], int):
                    chunk_size = args[1]
                else:
                    # ttnn.split only supports chunks of same size.
                    return None

                new_args = (args[0], chunk_size, split_dim)
                return g.call_function(ttnn.split, args=new_args)

            if node.target == torch.ops.aten._to_copy.default:
                # Keep it if the graph output uses this op
                target_users_ops = [user.target for user in node.users.keys()]
                if "output" in target_users_ops:
                    return None

                src_dtype, dst_dtype = node.args[0].meta["val"].dtype, kwargs["dtype"]

                # casting to bool type is equivalent to checking if the value is not 0
                if dst_dtype == torch.bool:
                    return g.call_function(ttnn.ne, args=(args[0], 0))

                try:
                    ttnn_dtype = torch_dtype_to_ttnn_dtype(dst_dtype)
                    return g.call_function(ttnn.typecast, args=(node.args[0], ttnn_dtype))
                except:
                    pass

                # Keep it if the cast is not supported
                if src_dtype in [torch.bool, torch.int32, torch.int64] and dst_dtype not in [
                    torch.int32,
                    torch.int64,
                    torch.bfloat16,
                ]:
                    return None

                # Update the dtype of the input node's metadata value to match the destination dtype
                if hasattr(node.args[0], "meta") and "val" in node.args[0].meta:
                    node.args[0].meta["val"] = node.args[0].meta["val"].to(dst_dtype)

                # Essentially remove this op
                return node.args[0]

            if node.target == torch.ops.aten.fill.Scalar:
                return g.call_function(ttnn.fill, args=args)

            if node.target in [torch.ops.aten.masked_fill.Scalar, torch.ops.aten.masked_fill.Tensor]:
                # aten.masked_fill is equivalent to the following:
                # masked_fill = (tensor * (ones - mask)) + (mask * full)

                tensor, mask, fill_value = args
                tensor_shape = get_shape(gm, tensor)
                mask_shape = get_shape(gm, mask)

                # check if divisible, otherwise skip
                np_tensor_shp = np.array(tensor_shape)
                np_mask_shp = np.array(mask_shape)
                if np.sum(np_tensor_shp % np_mask_shp) == 0:
                    # check if the fill_value is a constant
                    if isinstance(fill_value, torch.fx.node.Node) and fill_value.op == "get_attr":
                        with unset_fake_temporarily():
                            fill_value_attr = getattr(gm, fill_value.target)
                            if fill_value_attr.numel() == 1:
                                # extract single fill value
                                fill_value = fill_value_attr.item()
                            else:
                                # extract torch.tensor constant
                                fill_value = fill_value_attr.data

                    if isinstance(fill_value, (float,)):
                        full_kwargs = {"fill_value": fill_value, "device": TtnnDevice(), "layout": TtnnTileLayout()}
                        full = g.call_function(ttnn.full, (list(tensor_shape),), full_kwargs)
                    else:
                        # No cases for non-scalar fill value.
                        return None

                    mask_bcst = mask
                    if not np.all(np_tensor_shp == np_mask_shp):
                        multiplier = np_tensor_shp // np_mask_shp
                        mask_bcst = g.call_function(target_wrappers.repeat, args=(mask, multiplier.tolist()))

                    kwargs = {"dtype": TtnnBfloat16(), "layout": TtnnTileLayout(), "device": TtnnDevice()}
                    ones = g.call_function(ttnn.ones, (tensor_shape,), kwargs)
                    mask_flip = g.call_function(ttnn.subtract, (ones, mask_bcst))
                    tensor_masked = g.call_function(ttnn.multiply, (tensor, mask_flip))

                    full_masked = g.call_function(ttnn.multiply, (mask_bcst, full))
                    masked_fill = g.call_function(ttnn.add, (tensor_masked, full_masked))

                    return masked_fill
                else:
                    return None

            if node.target == torch.ops.aten.select.int:
                tensor, dim, start = args

                input_shape = get_shape(gm, args[0])
                output_shape = get_shape(gm, node)

                slice_start, slice_end = [0] * len(input_shape), list(input_shape)
                slice_start[dim] = (start + input_shape[dim]) % input_shape[dim]
                slice_end[dim] = slice_start[dim] + 1

                slice_tensor = g.call_function(ttnn.slice, (tensor, [*slice_start], [*slice_end]))

                return g.call_function(ttnn.reshape, args=(slice_tensor, list(output_shape)))

            if node.target == torch.ops.aten.cumsum.default:
                tensor, dim = args
                input_shape = get_shape(gm, tensor)
                rank = len(input_shape)
                if rank > 4:
                    return None
                dim = (dim + rank) % rank
                # Unsqueeze input tensor to 4D for cumsum
                # TODO(#367): Special case if dim is inner-most 2 dim. Unsqueeze (x, y) to (x, y, 1, 1) as cumsum currently only support N and C
                if (dim - rank) >= -2:
                    if rank <= 2:
                        input_4d_shape = (1,) * (2 - rank) + (*input_shape, 1, 1)
                    elif rank == 3 and dim == 1:
                        input_4d_shape = (*input_shape, 1)
                    else:
                        return None
                else:
                    input_4d_shape = (1,) * (4 - rank) + input_shape
                    dim += 4 - rank
                input_4d = g.call_function(ttnn.reshape, (tensor, input_4d_shape))
                output_4d = g.call_function(ttnn.moreh_cumsum, (input_4d, dim), kwargs)
                return g.call_function(ttnn.reshape, (output_4d, input_shape))

            if node.target == torch.ops.aten.cat.default:
                tensors = args[0]
                if len(tensors) == 1:
                    return tensors[0]
                dim = 0
                if len(args) > 1:
                    dim = args[1]
                rank = len(node.meta["val"].size())
                dim = (dim + rank) % rank
                layout = TtnnTileLayout() if rank == 4 else TtnnRowMajorLayout()

                tensor_list = []
                for tensor in tensors:
                    tensor_list.append(g.call_function(ttnn.to_layout, (tensor,), {"layout": layout}))

                return g.call_function(ttnn.concat, (tensor_list, dim))

            if node.target == torch.ops.aten.sum.default:
                sum_tensor = g.call_function(ttnn.sum, (args[0],))
                return g.call_function(torch.ops.aten.squeeze.default, args=(sum_tensor,))

            if node.target == torch.ops.aten.max_pool2d_with_indices.default:
                params = map_args_to_kwargs(
                    args,
                    (
                        (0, "input_tensor"),
                        (1, "kernel_size"),
                        (2, "stride"),
                        (3, "padding"),
                        (4, "dilation"),
                        (5, "ceil_mode"),
                    ),
                )
                input_tensor = params["input_tensor"]
                input_shape = input_tensor.meta["val"].size()
                batch_size, in_c, in_h, in_w = input_shape
                kernel_size = params["kernel_size"]
                stride = params.get("stride", kernel_size)
                padding = params.get("padding", (0, 0))
                dilation = params.get("dilation", (1, 1))
                ceil_mode = params.get("ceil_mode", False)
                if (
                    # # TODO: in_c must be 16 or a multiple of 32
                    (in_c != 16 and in_c % 32 != 0)
                    # TODO(#419): Currently fails with in_c < 16
                    or in_c < 16
                    # TODO(tt-metal#12099): Currently it doesn't return indices. Convert if only the value is used
                    or not is_getitem_0_only_user(node)
                ):
                    return None

                input_tensor = insert_ncx_to_nxc(g, input_tensor)
                # TODO(#423): reshape can be removed if (N, H, W, C) is supported
                input_tensor = g.call_function(ttnn.reshape, (input_tensor, (1, 1, batch_size * in_h * in_w, in_c)))
                # TODO(#418): max_pool2d fails with input tensors in tile layout for some shapes
                input_tensor = g.call_function(ttnn.to_layout, (input_tensor, TtnnRowMajorLayout()))
                output_tensor = g.call_function(
                    ttnn.max_pool2d,
                    (
                        input_tensor,
                        batch_size,
                        in_h,
                        in_w,
                        in_c,
                        kernel_size,
                        stride,
                        padding,
                        dilation,
                    ),
                    {"ceil_mode": ceil_mode},
                )
                output_tensor = insert_sharded_nxc_to_ncx(g, output_tensor, node.meta["val"][0].size())
                # TODO(tt-metal#12099): Currently it doesn't return indices. Pack into tuple to maintain the type
                return g.call_function(target_wrappers.pack_to_tuple, (output_tensor,))

            if node.target == torch.ops.aten.avg_pool2d.default:
                params = map_args_to_kwargs(
                    args,
                    (
                        (0, "input_tensor"),
                        (1, "kernel_size"),
                        (2, "stride"),
                        (3, "padding"),
                        (4, "ceil_mode"),
                        (5, "count_include_pad"),
                        (6, "divisor_override"),
                    ),
                )
                input_tensor = params["input_tensor"]
                input_shape = input_tensor.meta["val"].size()
                batch_size, in_c, in_h, in_w = input_shape
                kernel_size = params["kernel_size"]
                stride = params.get("stride", kernel_size)
                padding = params.get("padding", (0, 0))
                ceil_mode = params.get("ceil_mode", False)
                count_include_pad = params.get("count_include_pad", True)
                divisor_override = params.get("divisor_override", None)

                if (
                    # # TODO: in_c must be 16 or a multiple of 32
                    (in_c != 16 and in_c % 32 != 0)
                    # TODO(#419): Currently fails with in_c < 16
                    or in_c < 16
                ):
                    return None
                if ceil_mode or not count_include_pad or divisor_override is not None:
                    return None
                input_tensor = insert_ncx_to_nxc(g, input_tensor)
                input_tensor = g.call_function(ttnn.reshape, (input_tensor, (1, 1, batch_size * in_h * in_w, in_c)))
                input_tensor = g.call_function(ttnn.to_layout, (input_tensor, TtnnRowMajorLayout()))
                output_tensor = g.call_function(
                    ttnn.avg_pool2d,
                    (
                        input_tensor,
                        batch_size,
                        in_h,
                        in_w,
                        in_c,
                        kernel_size,
                        stride,
                        padding,
                    ),
                    (
                        {
                            "ceil_mode": ceil_mode,
                        }
                    ),
                )
                return insert_sharded_nxc_to_ncx(g, output_tensor, node.meta["val"].size())

            if node.target == torch.ops.aten._native_batch_norm_legit_no_training.default:
                return batch_norm_inference(*args)

            if node.target == torch.ops.aten.convolution.default:
                params = map_args_to_kwargs(
                    args,
                    (
                        (0, "input_tensor"),
                        (1, "weight_tensor"),
                        (2, "bias_tensor"),
                        (3, "stride"),
                        (4, "padding"),
                        (5, "dilation"),
                        (6, "transposed"),
                        (7, "output_padding"),
                        (8, "groups"),
                    ),
                )
                input_node = params["input_tensor"]
                input_shape = list(input_node.meta["val"].size())
                in_spatial_shape = input_shape[2:]
                in_spatial_size = len(in_spatial_shape)
                weight_tensor = params["weight_tensor"]
                weight_shape = list(weight_tensor.meta["val"].size())
                kernel_spatial_shape = weight_shape[2:]

                transposed = params.get("transposed", False)
                groups = params.get("groups", 1)
                stride = params.get("stride", [1] * in_spatial_size)
                padding = params.get("padding", [0] * in_spatial_size)
                dilation = params.get("dilation", [1] * in_spatial_size)
                output_padding = params.get("output_padding", [0] * in_spatial_size)

                # TODO(tt-metal#16258): conv1d API doesn't support transposed yet
                if in_spatial_size == 1 and transposed:
                    return None

                batch_size, in_c = input_shape[:2]
                in_nhw = math.prod([batch_size] + in_spatial_shape)
                out_c = weight_shape[1] if transposed else weight_shape[0]

                input_tensor = insert_ncx_to_nxc(g, input_node)
                # TODO(tt-metal#15148): ttnn.conv2d internal reshape fails with padding
                input_tensor = g.call_function(ttnn.reshape, (input_tensor, (1, 1, in_nhw, in_c)))
                bias_tensor = params.get("bias_tensor", None)
                if bias_tensor is not None:
                    bias_shape = bias_tensor.meta["val"].size()
                    bias_tensor = g.call_function(
                        ttnn.reshape,
                        (bias_tensor, (1,) * (4 - len(bias_shape)) + bias_shape),
                    )
                    bias_tensor = g.call_function(target_wrappers.move_to_host, (bias_tensor, TtnnRowMajorLayout()))
                output_tensor = g.call_function(
                    target_wrappers.conv,
                    (
                        input_tensor,
                        weight_tensor,
                        bias_tensor,
                        batch_size,
                        in_c,
                        out_c,
                        in_spatial_shape,
                        kernel_spatial_shape,
                        stride,
                        padding,
                        dilation,
                        groups,
                        TtnnDevice(),
                        transposed,
                        output_padding if transposed else None,
                    ),
                )
                return insert_sharded_nxc_to_ncx(g, output_tensor, node.meta["val"].size())

            if node.target == torch.ops.aten.slice_scatter.default:
                tensor, src_tensor, dim, start, end, *step = args

                tensor_shape = tensor.meta["val"].size()
                src_tensor_shape = src_tensor.meta["val"].size()

                # nothing to slice, result tensors == src_tensor
                if tensor_shape == src_tensor_shape:
                    return src_tensor

                # Convert both inputs to bfloat16 because ttnn.concat does not work with certain
                # shapes if they are of uint32 dtype.
                # https://github.com/tenstorrent/tt-metal/issues/20376
                if tensor.meta["val"].dtype not in [torch.bfloat16, torch.float]:
                    tensor = g.call_function(ttnn.typecast, (tensor, TtnnBfloat16()))

                if src_tensor.meta["val"].dtype not in [torch.bfloat16, torch.float]:
                    src_tensor = g.call_function(ttnn.typecast, (src_tensor, TtnnBfloat16()))

                # slice_scatter could be concat([pre_slice_tensor, src_tensor, post_slice_tensor])
                rank = len(tensor_shape)
                [step] = step or [1]

                if step != 1:
                    return None

                assert dim < rank, f"The slice dim {dim} should be less than rank {rank}"

                dim = (dim + rank) % rank
                start = start if start is not None else 0
                start = start if start >= 0 else (start + tensor_shape[dim])
                end = end if end is not None else tensor_shape[dim]
                end = end if end >= 0 else (end + tensor_shape[dim])
                end = 0 if end < 0 else min(tensor_shape[dim], end)

                tensors_list = []
                # pre_slice_tensor
                if start > 0:
                    slice_start = [0] * rank
                    slice_end = list(tensor_shape)
                    slice_end[dim] = start
                    tensors_list.append(g.call_function(ttnn.slice, (tensor, slice_start, slice_end)))

                # src_tensor
                tensors_list.append(src_tensor)

                # post_slice_tensor
                if end < tensor_shape[dim]:
                    slice_start = [0] * rank
                    slice_start[dim] = end
                    slice_end = list(tensor_shape)
                    tensors_list.append(g.call_function(ttnn.slice, (tensor, slice_start, slice_end)))

                # concat all together
                tensors_to_concat = []
                for tensor in tensors_list:
                    tensors_to_concat.append(g.call_function(ttnn.to_layout, (tensor,), {"layout": TtnnTileLayout()}))

                # Bug: https://github.com/tenstorrent/tt-metal/issues/20205
                # If rank is 1D, unsqueeze to 2D, then squeeze back to 1D
                # Delete this block once the bug is fixed.
                if rank == 1:
                    for idx, ten in enumerate(tensors_to_concat):
                        tensors_to_concat[idx] = g.call_function(ttnn.unsqueeze, (ten, 0))

                    concat = g.call_function(ttnn.concat, (tensors_to_concat, dim + 1))
                    return g.call_function(ttnn.squeeze, (concat, 0))

                return g.call_function(ttnn.concat, (tensors_to_concat, dim))

            if node.target == torch.ops.aten.argmax.default:
                tensor, *dim = args
                rank = len(tensor.meta["val"].size())
                tt_kwargs = {}
                if dim:
                    dim = (dim[0] + rank) % rank
                    tt_kwargs["dim"] = dim

                return g.call_function(ttnn.argmax, args=(tensor,), kwargs=tt_kwargs)

            if node.target == torch.ops.aten.stack.default:
                tensors, *dim = args
                dim = dim[0] if dim else 0
                output_shape = list(node.meta["val"].size())
                return g.call_function(target_wrappers.stack, (tensors, dim, output_shape))

            if node.target == torch.ops.aten.roll.default:
                tensor, shifts, dims = args
                input_shape = list(tensor.meta["val"].size())
                return g.call_function(target_wrappers.roll, (tensor, input_shape, shifts, dims))

            if node.target == torch.ops.aten.bitwise_not.default:
                input_type = get_dtype(args[0])
                if input_type == torch.bool:
                    return g.call_function(ttnn.eq, args=(args[0], 0))

                # ttnn.bitwise_not only supports int32
                cast_to_int32 = g.call_function(ttnn.typecast, args=(args[0], TtnnInt32()))
                return g.call_function(ttnn.bitwise_not, args=(cast_to_int32,))

            if node.target == torch.ops.aten.all.default:
                input_shape = get_shape(gm, args[0])
                ttnn_all = g.call_function(target_wrappers.all, args=(args[0], input_shape.numel()))
                return g.call_function(torch.ops.aten.squeeze.default, args=(ttnn_all,))

            if node.target == torch.ops.aten.copy.default:
                # Equivalent to in-place torch.copy_.
                return args[1]

            if node.target == torch.ops.aten.empty.memory_format:
                # raise RuntimeError(f"{str(kwargs)}, {str(args)}, {str(type(args[0]))}")
                dtype_mapping = {
                    torch.float32: TtnnBfloat16(),
                    torch.float16: TtnnBfloat16(),
                    torch.int32: TtnnInt32(),
                }
                dtype = dtype_mapping.get(kwargs["dtype"], TtnnUint32())
                new_kwargs = {
                    "dtype": dtype,
                    "layout": TtnnTileLayout(),
                    "device": TtnnDevice(),
                }
                return g.call_function(ttnn.empty, args=(args[0],), kwargs=new_kwargs)
            if node.target == torch.ops.aten._scaled_dot_product_flash_attention.default:

                def pad_qkv_ttnn(q, k, v, scale=None, align_by=ttnn.TILE_SIZE):
                    """
                    Pads the last dimension of Q, K, V tensors to `tile_size`
                    and rescales Q so that the scaled-dot-product attention
                    is numerically identical to the unpadded case.
                    """
                    # Extract shape from metadata
                    q_shape = list(q.meta["val"].shape)
                    d = q_shape[-1]
                    pad = (-d) % align_by  # 0 – 31

                    padded_shape = q_shape.copy()
                    padded_shape[-1] = d + pad

                    if pad:
                        ttnn_pad = [(0, pad)]

                        # metadata is broken because torch fx expects sdpa to return a tuple
                        torch_pad = tuple(x for pair in ttnn_pad for x in pair)
                        meta_val = torch.nn.functional.pad(g._get_val(q), torch_pad, value=0)

                        # 1) zero‑pad along head_dim
                        q = g.call_function(ttnn.pad, args=(q, ttnn_pad, 0))
                        k = g.call_function(ttnn.pad, args=(k, ttnn_pad, 0))
                        v = g.call_function(ttnn.pad, args=(v, ttnn_pad, 0))

                        q.meta["val"] = meta_val
                        k.meta["val"] = meta_val
                        v.meta["val"] = meta_val

                        # 2) rescale Q so that Q*K^T / sqrt(d') == Q_p*K_p^T / sqrt(d)
                        if scale is None:
                            calc_scale = math.sqrt((d + pad) / d)
                            q = g.call_function(ttnn.multiply, args=(q, calc_scale))

                    return q, k, v, d

                def select(dropout_p=0.0, is_causal=False):
                    # TODO(jdh8): Add support for training mode
                    if dropout_p > 0.0 or not is_getitem_0_only_user(node):
                        return None

                    # Pad last dimension of Q, K, V to tile size
                    q, k, v = args[:3]
                    q_padded, k_padded, v_padded, d = pad_qkv_ttnn(q, k, v, scale=kwargs.get("scale", None))

                    # TODO: Those configs can be further optimized based on input shape to get
                    # best performance/accuracy tradeoff
                    default_q_chunk_size = 32  # <- The bigger the better accuracy but slower
                    default_k_chunk_size = 32

                    compute_config_params = {
                        "math_fidelity": ttnn.MathFidelity.HiFi2,
                        "math_approx_mode": True,
                        "fp32_dest_acc_en": False,
                        "packer_l1_acc": False,
                    }
                    if ttnn.device.is_wormhole_b0(device):
                        compute_kernel_config = get_emplace_custom_object_in_graph(
                            ttnn.WormholeComputeKernelConfig, **compute_config_params
                        )
                    elif ttnn.device.is_blackhole(device):
                        compute_kernel_config = get_emplace_custom_object_in_graph(
                            ttnn.BlackholeComputeKernelConfig, **compute_config_params
                        )
                    else:
                        raise RuntimeError(f"Unsupported device for {device.arch()} compute kernel config")
                    program_config = get_emplace_custom_object_in_graph(
                        ttnn.SDPAProgramConfig,
                        compute_with_storage_grid_size=device.compute_with_storage_grid_size(),
                        q_chunk_size=default_q_chunk_size,
                        k_chunk_size=default_k_chunk_size,
                        exp_approx_mode=False,
                    )
                    res_node = g.call_function(
                        ttnn.transformer.scaled_dot_product_attention,
                        args=(q_padded, k_padded, v_padded),
                        kwargs={
                            "is_causal": is_causal,
                            "scale": kwargs.get("scale", None),
                            "compute_kernel_config": compute_kernel_config,
                            "program_config": program_config,
                        },
                    )

                    # If padding was applied, slice the result back to original dimension
                    pad = (-d) % 32
                    if pad:
                        output_shape = list(res_node.meta["val"][0].shape)
                        slice_start = [0] * len(output_shape)
                        slice_end = output_shape.copy()
                        slice_end[-1] = d  # Set last dimension back to original size
                        res_node = g.call_function(ttnn.slice, (res_node, slice_start, slice_end))

                    # torch.ops.aten._scaled_dot_product_flash_attention.default return a tuple of values and inserts a
                    # getitem(ret, 0) after it. ttnn.transformer.scaled_dot_product_attention only returns one value.
                    if (val := res_node.meta.get("val", None)) is not None:
                        res_node.meta["val"] = val[0]

                    return res_node

                return select(*args[3:])

            # Removes getitem after ttnn.transformer.scaled_dot_product_attention
            # Related to https://github.com/tenstorrent/tt-metal/issues/16021
            if node.target == operator.getitem and isinstance(args[1], int) and args[1] == 0:
                if args[0].target == ttnn.transformer.scaled_dot_product_attention or (
                    args[0].target == ttnn.slice
                    and args[0].args[0].target == ttnn.transformer.scaled_dot_product_attention
                ):
                    return args[0]

            # PEP 8 suggests this explicit statement
            return None

        modified = False
        with g.inserting_before(node):
            new_node = rewrite_node(node)
            if new_node is not None:
                modified = True
                node.replace_all_uses_with(
                    new_node,
                    delete_user_cb=lambda node: node != new_node,
                )

    gm = GraphCleanup(gm)
    return gm, modified


def decompose_aten_to_aten_ops(gm: torch.fx.GraphModule, g: GraphWrapper, node):
    args = node.args
    kwargs = node.kwargs
    if node.target == torch.ops.aten.full_like.default:
        target_shape = args[0].meta["val"].size()
        return g.call_function(torch.ops.aten.full.default, args=(target_shape, *args[1:]), kwargs=kwargs)

    if node.target == torch.ops.aten.new_full.default:
        target_shape = args[1]
        new_kwargs = dict(kwargs)
        # Use the inferred output dtype so we don't need to figure out the dtype by ourselves
        new_kwargs["dtype"] = node.meta["val"].dtype
        return g.call_function(torch.ops.aten.full.default, args=(target_shape, *args[2:]), kwargs=new_kwargs)

    if node.target == torch.ops.aten.new_zeros.default:
        target_shape = args[1]
        new_kwargs = dict(kwargs)
        # Use the inferred output dtype so we don't need to figure out the dtype by ourselves
        new_kwargs["dtype"] = node.meta["val"].dtype
        return g.call_function(torch.ops.aten.zeros.default, args=(target_shape, *args[2:]), kwargs=new_kwargs)

    if node.target == torch.ops.aten._log_softmax.default:
        dim, half_to_float = get_arg(node, 1, "dim"), get_arg(node, 2, "half_to_float")
        softmax = g.call_function(torch.ops.aten._softmax.default, args=(args[0], dim, half_to_float))
        log = g.call_function(torch.ops.aten.log.default, args=(softmax,))
        return log

    if node.target == torch.ops.aten.clamp_min.default:
        return g.call_function(torch.ops.aten.clamp.default, args=(args[0], args[1], None))

    if node.target == torch.ops.aten.select.int:
        input_shape = get_shape(gm, args[0])
        output_shape = get_shape(gm, node)
        if input_shape.numel() == output_shape.numel() and len(output_shape) == 0:
            return g.call_function(torch.ops.aten.squeeze.dim, args=(args[0], 0))
        return None

    if node.target == torch.ops.aten.sum.default:
        input_shape = get_shape(gm, args[0])
        output_shape = get_shape(gm, node)
        if input_shape.numel() == 1:
            return g.call_function(torch.ops.aten.squeeze.default, args=(args[0],))
        return None

    if node.target in [torch.ops.aten.index.Tensor, torch.ops.aten._unsafe_index.Tensor]:
        input_shape = get_shape(gm, args[0])
        indices = get_arg(node, 1, "indices")
        if len(input_shape) == 2 and len(indices) == 1 and indices[0] is not None:
            index_shape = get_shape(gm, indices[0])
            # magic number, 38000 can pass, 39000 can pass, but 38809 will hang
            # and if device is just ttnn.open_device(device=0), then it can pass
            # see issue #685
            if index_shape == torch.Size([38809]):
                return None
            return g.call_function(torch.ops.aten.embedding.default, args=(args[0], indices[0]))
        return None

    if node.target == torch.ops.aten.index_select.default:
        dim = get_arg(node, 1, "dim")
        indices = get_arg(node, 2, "indices")
        new_indices = [None] * dim + [indices]
        return g.call_function(torch.ops.aten.index.Tensor, args=(args[0], new_indices))
    return None


# TODO(jerrysky3): Refactor ReplaceMoreTtManually with rewrite_graph
def rewrite_graph(gm: torch.fx.GraphModule, rewrite_node_fn) -> torch.fx.GraphModule:
    nodes = list(gm.graph.nodes)
    modified = False
    for node in nodes:
        if not can_lowering_to_ttnn(node):
            continue
        g = GraphWrapper(gm, node)
        with g.inserting_before(node):
            new_node = rewrite_node_fn(gm, g, node)
            if new_node is not None:
                modified = True
                node.replace_all_uses_with(
                    new_node,
                    delete_user_cb=lambda node: node != new_node,
                )

    gm = GraphCleanup(gm)
    return gm, modified


class ToTtPass(PassBase):
    """Pass to convert aten ops to ttnn ops.

    This pass is currently multi-device aware since it rewrites the shard / replicate / concat operations inserted by the MultiDevicePass. Some aten to ttnn conversions are performed with the torch.fx.Transformer, while others are manually rewritten.

    :param device: The device on which the workflow will execute. Can be a MeshDevice or Device.
    :param use_less_ttnn_op_types: Whether to use less ttnn op types (maybe unused).
    """

    def __init__(self, device, use_less_ttnn_op_types):
        self.device = device
        self.use_less_ttnn_op_types = use_less_ttnn_op_types

    def call(self, gm: torch.fx.GraphModule):
        # Decompose some aten ops to simpler aten ops
        max_try = 10
        cnt = 0
        while True:
            cnt += 1
            gm, modified = rewrite_graph(gm, decompose_aten_to_aten_ops)
            if not modified:
                break
            if cnt == max_try:
                raise RuntimeError("Failed to decompose aten ops to simpler aten ops")

        # Replace more patterns with torch.fx.Transformer
        gm = ReplaceMoreTt(gm, self.device, self.use_less_ttnn_op_types).transform()

        # Replace patterns manually
        max_try = 10
        cnt = 0
        while True:
            cnt += 1
            gm, modified = ReplaceMoreTtManually(gm, self.device, self.use_less_ttnn_op_types)
            if not modified:
                break
            if cnt == max_try:
                raise RuntimeError("Failed to decompose aten ops to simpler aten ops")

        return PassResult(gm, True)



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
from typing import List
import torch
import pickle
from pathlib import Path
from torch._subclasses import FakeTensor
from torch.fx.experimental.symbolic_shapes import hint_int


# Save a pickle file from a Python object to metrics/{base_path}/{filename}.pickle
def save_pickle(obj, base_path, filename):
    p = Path(f"metrics/{base_path}")
    p.mkdir(parents=True, exist_ok=True)
    pickle_out_path = p / f"{filename}.pickle"
    with open(pickle_out_path, "wb") as f:
        pickle.dump(obj, f)


def _get_shape_from_fake_tensor(faketensor):
    if isinstance(faketensor, torch._subclasses.fake_tensor.FakeTensor):
        return list(faketensor.size())
    else:
        return faketensor


def _sanitize_value(value):
    if isinstance(value, list):
        val_list = []
        for val in value:
            if isinstance(val, torch.fx.proxy.Proxy):
                val = val.node
            if isinstance(val, torch.fx.node.Node) and "val" in val.meta:
                val_shape = _get_shape_from_fake_tensor(val.meta["val"])
                val_list.append(f"<{val_shape}>")
            else:
                val_list.append(str(val))
        return f"[{', '.join(val_list)}]"
    else:
        return value


class Inputs:
    def __init__(self, dtype, name, shape, value):
        self.dtype = dtype
        self.name = name
        self.shape_ = shape
        self.value_ = value

        self.shape = f"<{_get_shape_from_fake_tensor(shape)}>" if shape is not None else ""
        self.value = f"{str(_sanitize_value(value))}" if value is not None else "?"

    def __repr__(self):
        return f"{str(self.dtype)}{self.shape} {str(self.name)} = {self.value}"


class InputVariation:
    def __init__(self, opname, inputs: List[Inputs]):
        self.opname = opname
        self.inputs = inputs

    def get_input_str_list(self):
        return [str(input) for input in self.inputs]

    def __repr__(self):
        input_str = ", ".join(self.get_input_str_list())
        return f"{self.opname}:\n\t{input_str}\n"

    def dict(self):
        return {"opname": str(self.opname), "inputs": self.get_input_str_list()}

    def dict_for_pickle(self):
        out = {"opname": str(self.opname)}

        inputs = []
        tensor_infos = []
        for obj in self.inputs:
            inputs.append(str(obj))
            _obj = obj.shape_
            if isinstance(_obj, FakeTensor):
                info = {}
                info["name"] = str(obj)
                info["data_type"] = str(_obj.dtype)
                info["buffer_type"] = "default"
                info["layout"] = str(_obj.layout)
                info["grid_shape"] = []

                shape = []
                for _s in _obj.shape:
                    if type(_s) is torch.SymInt:
                        shape.append(hint_int(_s))
                        continue

                    shape.append(_s)

                info["shape"] = shape

                tensor_infos.append(info)

        out["inputs"] = inputs
        out["tensor_infos"] = tensor_infos

        return out


class ConvertedInput:
    def __init__(self, opname, original_input_variation: InputVariation):
        self.opname = opname
        self.original_input_variation = original_input_variation

    def __repr__(self):
        return f"{self.opname} --> {self.original_input_variation}"

    def dict(self):
        return {"opname": str(self.opname), "original_inputs": self.original_input_variation.dict()}


def collect_input_variation(target, args, kwargs):
    """
    Collect the input variation from schema of a target function if exists

    Returns:
        InputVariation class or None
    """
    if hasattr(target, "_schema"):
        # Get schema list for this op
        schema_pos_args = [arg for arg in target._schema.arguments if not arg.kwarg_only]
        schema_kw_args = {arg.name: arg.type for arg in target._schema.arguments if arg.kwarg_only}

        # Map args with schema
        inputs = []
        for i, arg in enumerate(args):
            dtype = schema_pos_args[i].type
            name = schema_pos_args[i].name

            if isinstance(arg, torch.fx.proxy.Proxy):
                arg = arg.node

            # Collect the input shapes from the metadata if possible.
            if hasattr(arg, "meta") and "val" in arg.meta:
                shape = arg.meta["val"]
            else:
                shape = None

            # Collect the input values.
            # Unknown values will be substituted with an empty string.
            if not isinstance(arg, torch.fx.node.Node):
                value = arg
            else:
                value = None

            inputs.append(Inputs(dtype, name, shape, value))

        # Map kwargs with schema
        for name, value in kwargs.items():
            dtype = schema_kw_args[name]
            # Collect the shape of a value if it's a node
            if isinstance(value, torch.fx.node.Node):
                shape = value.meta["val"]
            else:
                shape = None
            inputs.append(Inputs(dtype, name, shape, value))

        return InputVariation(target, inputs)
    else:
        return None


def collect_input_variation_from_node(node: torch.fx.Node):
    """
    Collect input variation from a torch.fx.Node. If the node already has an original
    input variation attached, then create a ConvertedInput.

    TTNN Data-movement ops are not considered converted at the moment.

    Returns:
        class InputVariation or class ConvertedInput
    """
    if hasattr(node, "meta") and ((original_input_variations := node.meta.get("original_input_variations")) != None):
        # aten ops should begin with "aten."
        # however, some ttnn ops are wrapped, so they will be missing the prefix
        opname = str(node.target) if str(node.target).startswith("aten.") else node.target.__name__
        if isinstance(original_input_variations, ConvertedInput):
            return ConvertedInput(opname, original_input_variations.original_input_variation)

        assert isinstance(original_input_variations, InputVariation)
        if original_input_variations.opname != opname:
            return ConvertedInput(opname, original_input_variations)
    return collect_input_variation(node.target, node.args, node.kwargs)


def collect_input_variations_from_list_nodes(nodes: List[torch.fx.Node]):
    """
    Collect all the input variations from a list of torch.fx.Nodes.

    Returns:
        List of InputVariation or ConvertedInput class objects.
    """
    collection = []
    for node in nodes:
        if (input_variations := collect_input_variation_from_node(node)) is not None:
            collection.append(input_variations)

    return collection



# SPDX-FileCopyrightText: © 2025 Tenstorrent AI ULC
#
# SPDX-License-Identifier: Apache-2.0
import torch
import ttnn
from torch.fx.passes.infra.pass_base import PassBase, PassResult
from torch_ttnn.passes.lowering.add_data_move_pass import is_tt_data_move, is_tt_compute

from torch_ttnn.mem_utils import *


def evict_tensors_from_device(
    gm: torch.fx.GraphModule, mm: MemoryManager, guilty_op: str, tensors_to_evict: list
) -> torch.fx.GraphModule:
    assert len(tensors_to_evict) != 0, "Tensors to be evicted list must not be empty."
    topo_sorted_nodelist = list(gm.graph.nodes)
    print(topo_sorted_nodelist)

    for node in topo_sorted_nodelist:
        # The node whose output tensor needs to be evicted
        tid = -1  # For init
        if node in mm.node_to_tid_map:
            tid = mm.node_to_tid_map[node]
        if tid in tensors_to_evict:
            assert (
                node.name != guilty_op
            ), "Op which overflows the SRAM can't have its output tensor evicted from device"
            evict_node = node
            evict_node_users = list(evict_node.users.keys())

        g = node.graph

        if is_tt_compute(node) and node.name == guilty_op:
            with g.inserting_before(node):
                from_device_node = g.call_function(ttnn.from_device, args=(evict_node,))

            with g.inserting_after(node):
                device = TtnnDevice()
                to_device_node = g.call_function(ttnn.to_device, (from_device_node, device))

            assert evict_node is not None and evict_node_users != [], "Evicted node should be there."
            # For users of evict node, if it topologically comes after the evict node:
            # then update the user's argument (input) to be "to_device node"
            for user in evict_node_users:
                if topo_sorted_nodelist.index(user) > topo_sorted_nodelist.index(node):
                    for input in user.all_input_nodes:
                        if input == evict_node:
                            index = user.all_input_nodes.index(input)
                            user.update_arg(index, to_device_node)
            evict_node = None
            evict_node_users = []

    print(list(gm.graph.nodes))
    return gm


class EvictionPass(PassBase):
    def __init__(self, mm, guilty_op, tensors_to_evict):
        self.mm = mm
        self.guilty_op = guilty_op
        self.tensors_to_evict = tensors_to_evict

    def call(self, gm: torch.fx.GraphModule):
        gm = evict_tensors_from_device(gm, self.mm, self.guilty_op, self.tensors_to_evict)
        return PassResult(gm, True)
